from .base_product import SpeasyProduct
from .catalog import Catalog, Event
from .timetable import TimeTable
from .dataset import Dataset
from .variable import SpeasyVariable, VariableTimeAxis, VariableAxis, DataContainer
from typing import Optional, Union, List

MaybeAnyProduct = Optional[Union[SpeasyProduct, List[SpeasyProduct]]]
MaybeTimeDependentProduct = Optional[Union[SpeasyVariable, Dataset]]
MaybeTimeIndependentProduct = Optional[Union[TimeTable, Catalog]]

__all__ = ['SpeasyVariable', 'Catalog', 'Event', 'Dataset', 'TimeTable', 'MaybeAnyProduct', 'MaybeTimeDependentProduct',
           'MaybeTimeIndependentProduct', 'VariableAxis', 'VariableTimeAxis', 'DataContainer']



from datetime import datetime
from sys import getsizeof
from typing import Dict, List

import astropy.units
import numpy as np


def _to_index(key, time):
    if key is None:
        return None
    if type(key) in (int, np.int64, np.int32, np.uint64, np.uint32):
        return key
    if isinstance(key, float):
        return np.searchsorted(time, np.datetime64(int(key * 1e9), 'ns'), side='left')
    if isinstance(key, datetime):
        return np.searchsorted(time, np.datetime64(key, 'ns'), side='left')
    if isinstance(key, np.datetime64):
        return np.searchsorted(time, key, side='left')


class DataContainer(object):
    __slots__ = ['__values', '__name', '__meta', '__is_time_dependent']

    def __init__(self, values: np.array, meta: Dict = None, name: str = None, is_time_dependent: bool = True):
        self.__values = values
        self.__is_time_dependent = is_time_dependent
        self.__name = name or ""
        self.__meta = meta or {}

    def reshape(self, new_shape):
        self.__values = self.__values.reshape(new_shape)

    @property
    def is_time_dependent(self) -> bool:
        return self.__is_time_dependent

    @property
    def values(self) -> np.array:
        return self.__values

    @property
    def shape(self):
        return self.__values.shape

    @property
    def unit(self) -> str:
        return self.__meta.get('UNITS')

    @property
    def nbytes(self) -> int:
        return self.__values.nbytes + getsizeof(self.__meta) + getsizeof(self.__name)

    def view(self, index_range: slice):
        return DataContainer(name=self.__name, meta=self.__meta, values=self.__values[index_range],
                             is_time_dependent=self.__is_time_dependent)

    def unit_applied(self, unit: str or None = None) -> "DataContainer":
        try:
            u = astropy.units.Unit(unit or self.unit)
        except (ValueError, KeyError):
            u = astropy.units.Unit("")

        return DataContainer(values=self.__values * u, meta=self.__meta, name=self.__name,
                             is_time_dependent=self.__is_time_dependent)

    def to_dictionary(self, array_to_list=False) -> Dict[str, object]:
        return {
            "values": self.__values.tolist() if array_to_list else self.__values.copy(),
            "meta": self.__meta.copy(),
            "name": self.__name,
            "is_time_dependent": self.is_time_dependent
        }

    @staticmethod
    def from_dictionary(dictionary: Dict[str, str or Dict[str, str] or List], dtype=np.float64) -> "DataContainer":
        try:
            return DataContainer(values=np.array(dictionary["values"], dtype=dtype), meta=dictionary["meta"],
                                 name=dictionary["name"],
                                 is_time_dependent=dictionary["is_time_dependent"])
        except ValueError:
            return DataContainer(values=np.array(dictionary["values"]), meta=dictionary["meta"],
                                 name=dictionary["name"],
                                 is_time_dependent=dictionary["is_time_dependent"])

    @staticmethod
    def reserve_like(other: 'DataContainer', length: int = 0) -> 'DataContainer':
        return DataContainer(name=other.__name, meta=other.__meta,
                             values=np.empty(
                                 (length,) + other.shape[1:], dtype=other.__values.dtype),
                             is_time_dependent=other.__is_time_dependent
                             )

    def __len__(self):
        return len(self.__values)

    def __getitem__(self, key):
        return self.view(key)

    def __setitem__(self, k, v: 'DataContainer'):
        assert type(v) is DataContainer
        self.__values[k] = v.__values

    def __eq__(self, other: 'DataContainer') -> bool:
        return self.__meta == other.__meta and \
               self.__name == other.__name and \
               self.is_time_dependent == other.is_time_dependent and \
               np.all(self.__values.shape == other.__values.shape) and \
               np.array_equal(self.__values, other.__values, equal_nan=True)

    def replace_val_by_nan(self, val):
        if self.__values.dtype != np.float64:
            self.__values = self.__values.astype(np.float64)
        self.__values[self.__values == val] = np.nan

    @property
    def meta(self):
        return self.__meta

    @property
    def name(self):
        return self.__name


class VariableAxis(object):
    __slots__ = ['__data']

    def __init__(self, values: np.array = None, meta: Dict = None, name: str = "", is_time_dependent: bool = False,
                 data: DataContainer = None):
        if data is not None:
            self.__data = data
        else:
            self.__data = DataContainer(
                values=values, name=name, meta=meta, is_time_dependent=is_time_dependent)

    def to_dictionary(self, array_to_list=False) -> Dict[str, object]:
        d = self.__data.to_dictionary(array_to_list=array_to_list)
        d.update({"type": "VariableAxis"})
        return d

    @staticmethod
    def from_dictionary(dictionary: Dict[str, str or Dict[str, str] or List], time=None) -> "VariableAxis":
        assert dictionary['type'] == "VariableAxis"
        return VariableAxis(data=DataContainer.from_dictionary(dictionary))

    @staticmethod
    def reserve_like(other: 'VariableAxis', length: int = 0) -> 'VariableAxis':
        return VariableAxis(data=DataContainer.reserve_like(other.__data, length))

    def __getitem__(self, key):
        if isinstance(key, slice):
            return self.view(slice(_to_index(key.start, self.__data.values), _to_index(key.stop, self.__data.values)))

    def __setitem__(self, k, v: 'VariableAxis'):
        assert type(v) is VariableAxis
        self.__data[k] = v.__data

    def __len__(self):
        return len(self.__data)

    def view(self, index_range: slice) -> 'VariableAxis':
        return VariableAxis(data=self.__data[index_range])

    def __eq__(self, other: 'VariableAxis') -> bool:
        return type(other) is VariableAxis and self.__data == other.__data

    @property
    def unit(self) -> str:
        return self.__data.unit

    @property
    def is_time_dependent(self) -> bool:
        return self.__data.is_time_dependent

    @property
    def values(self) -> np.array:
        return self.__data.values

    @property
    def shape(self):
        return self.__data.shape

    @property
    def name(self) -> str:
        return self.__data.name

    @property
    def nbytes(self) -> int:
        return self.__data.nbytes


class VariableTimeAxis(object):
    __slots__ = ['__data']

    def __init__(self, values: np.array = None, meta: Dict = None, data: DataContainer = None):
        if data is not None:
            self.__data = data
        else:
            if values.dtype != np.dtype('datetime64[ns]'):
                raise ValueError(
                    f"Please provide datetime64[ns] for time axis, got {values.dtype}")
            self.__data = DataContainer(
                values=values, name='time', meta=meta, is_time_dependent=True)

    def to_dictionary(self, array_to_list=False) -> Dict[str, object]:
        d = self.__data.to_dictionary(array_to_list=array_to_list)
        d.update({"type": "VariableTimeAxis"})
        return d

    @property
    def shape(self):
        return self.__data.shape

    @staticmethod
    def from_dictionary(dictionary: Dict[str, str or Dict[str, str] or List], time=None) -> "VariableTimeAxis":
        assert dictionary['type'] == "VariableTimeAxis"
        return VariableTimeAxis(data=DataContainer.from_dictionary(dictionary, dtype=np.dtype('datetime64[ns]')))

    @staticmethod
    def reserve_like(other: 'VariableTimeAxis', length: int = 0) -> 'VariableTimeAxis':
        return VariableTimeAxis(data=DataContainer.reserve_like(other.__data, length))

    def __getitem__(self, key):
        if isinstance(key, slice):
            return self.view(key)

    def __setitem__(self, k, v: 'VariableTimeAxis'):
        assert type(v) is VariableTimeAxis
        self.__data[k] = v.__data

    def __len__(self):
        return len(self.__data)

    def view(self, index_range: slice) -> "VariableTimeAxis":
        return VariableTimeAxis(data=self.__data[index_range])

    def __eq__(self, other: 'VariableTimeAxis') -> bool:
        return type(other) is VariableTimeAxis and self.__data == other.__data

    @property
    def is_time_dependent(self) -> bool:
        return True

    @property
    def values(self) -> np.array:
        return self.__data.values

    @property
    def unit(self) -> str:
        return 'ns'

    @property
    def name(self) -> str:
        return self.__data.name

    @property
    def nbytes(self) -> int:
        return self.__data.nbytes



""" catalog module
"""

from speasy.core.datetime_range import DateTimeRange
from .base_product import SpeasyProduct
from datetime import datetime
from typing import List
from speasy.core import all_of_type, listify
import pandas as pds


def _all_are_events(event_list):
    return all_of_type(event_list, Event)


class Event(DateTimeRange):
    """The Event class is a DatetimeRange with some meta data. It is supposed to be used with Catalog

    Attributes
    ----------
    start_time : datetime.datetime
    stop_time : datetime.datetime
    meta : dict
            Additional event data

    Methods
    -------

    Notes
    -----
    This class support the same operations as a speasy.common.datetime_range.DateTimeRange.

    See Also
    --------
    speasy.products.catalog.Catalog
    speasy.core.datetime_range.DateTimeRange
    """
    __slots__ = ['meta']

    def __init__(self, start_time: datetime, stop_time: datetime, meta=None):
        self.meta = meta or {}
        super().__init__(start_time, stop_time)

    def __eq__(self, other):
        return (self.meta == other.meta) and super().__eq__(other)

    def __repr__(self):
        return f"<Event: {self.start_time.isoformat()} -> {self.stop_time.isoformat()} | {self.meta}>"


class Catalog(SpeasyProduct):
    """The Catalog class allows to manipulate a goup of events like a simple Python list of Event plus some meta data.

    Attributes
    ----------
    name : str
        Catalog name
    meta : dict
        All additional Catalog meta data

    Methods
    -------
    append:
        Append an Event or a list of Event to the end of the Catalog
    pop:
        Remove and return Event at index (default last)

    Examples
    --------

    >>> import speasy
    >>> from speasy.products import Catalog, Event
    >>> my_catalog = Catalog(name='MyCatalog', meta={'tags':['demo', 'docstrings']}, events=[])
    >>> my_catalog.append(Event('2018-01-01', '2018-01-02', meta={'name':'My first event!'}))
    >>> my_catalog += Event('2019-01-01', '2019-01-02', meta={'name':'My second event!'})
    >>> for e in my_catalog:
    ...     print(e)
    ...
    <Event: 2018-01-01T00:00:00+00:00 -> 2018-01-02T00:00:00+00:00 | {'name': 'My first event!'}>
    <Event: 2019-01-01T00:00:00+00:00 -> 2019-01-02T00:00:00+00:00 | {'name': 'My second event!'}>

    See Also
    --------
    speasy.products.catalog.Event
    speasy.products.timetable.TimeTable
    """
    __slots__ = ['_events', 'name', 'meta']

    def __init__(self, name: str, meta: dict = None, events: List[Event] = None):
        super().__init__()
        self.name = name
        self.meta = meta or {}
        self._events = []
        if events:
            self.append(events)

    def __getitem__(self, index) -> Event:
        return self._events[index]

    def __len__(self) -> int:
        return len(self._events)

    def __iter__(self):
        return self._events.__iter__()

    def append(self, events: Event or List[Event]) -> None:
        """Append an Event or a list of Event to the end of the Catalog.

        Parameters
        ----------
        events : Event or List[Event]

        Raises
        ------
        TypeError
            If events is neither an Event or a list of Event

        See Also
        --------
        Catalog.pop
        """
        events = listify(events)
        if not _all_are_events(events):
            raise TypeError(
                f"You must provide a {Event} or a List of {Event} instead of {type(events)}")
        self._events += events

    def __iadd__(self, other: Event or List[Event]):
        self.append(other)
        return self

    def pop(self, index: int = -1) -> Event:
        """Remove and return Event at index (default last).

        Parameters
        ----------
        index : int

        Returns
        -------
        Event
            The removed event

        Raises
        ------
        IndexError
            if list is empty or index is out of range.

        """
        return self._events.pop(index)

    def to_dataframe(self) -> pds.DataFrame:
        columns = set()
        data = []
        for e in self:
            columns.update(e.meta.keys())
        columns = list(columns)
        for e in self:
            row = [e.start_time, e.stop_time] + [e.meta.get(column, None) for column in columns]
            data.append(row)
        columns = ['start_time', 'stop_time'] + columns
        return pds.DataFrame(columns=columns, data=data)

    def __repr__(self) -> str:
        return f"""<Catalog: {self.name}>"""



import logging
import pickle
from functools import wraps
from datetime import datetime, timedelta
import warnings

from packaging.version import Version
from dateutil import parser

from speasy.config import proxy as proxy_cfg
from speasy.config import inventories as inventories_cfg
from ... import SpeasyIndex

from ...products.variable import from_dictionary as var_from_dict
from .. import http
from ..inventory.indexes import from_dict as inventory_from_dict
from ..index import index

log = logging.getLogger(__name__)
PROXY_ALLOWED_KWARGS = ['disable_proxy']
MINIMUM_REQUIRED_PROXY_VERSION = Version("0.6.0")
_CURRENT_PROXY_SERVER_VERSION = None

if proxy_cfg.url() == "" or proxy_cfg.enabled() == False:
    warnings.warn("""Proxy server is disabled you might want to use it both to improve Speasy performances and to reduce pressure on remote servers
use the following python snippet to configure proxy server:
===========================================================================
import speasy as spz
spz.config.proxy.url.set("http://sciqlop.lpp.polytechnique.fr/cache")
spz.config.proxy.enabled.set(True)
===========================================================================
            """, stacklevel=0)


def query_proxy_version():
    global _CURRENT_PROXY_SERVER_VERSION
    if _CURRENT_PROXY_SERVER_VERSION is None:
        url = proxy_cfg.url()
        if url != "":
            resp = http.get(f"{url}/get_version?")
            if resp.status_code == 200:
                _CURRENT_PROXY_SERVER_VERSION = Version(resp.text.strip())
                return _CURRENT_PROXY_SERVER_VERSION
    return _CURRENT_PROXY_SERVER_VERSION


try:
    import zstd

    zstd_compression = 'true'


    def decompress(data):
        return zstd.decompress(data)

except ImportError:
    zstd_compression = 'false'


    def decompress(data):
        return data


class GetProduct:
    def __init__(self):
        pass

    @staticmethod
    def get(path: str, start_time: str, stop_time: str, **kwargs):
        url = proxy_cfg.url()
        kwargs['path'] = path
        kwargs['start_time'] = start_time
        kwargs['stop_time'] = stop_time
        kwargs['format'] = 'python_dict'
        kwargs['zstd_compression'] = zstd_compression
        resp = http.get(f"{url}/get_data?", params=kwargs)
        log.debug(f"Asking data from proxy {resp.url}, {resp.request.headers}")
        if resp.status_code == 200:
            var = var_from_dict(pickle.loads(decompress(resp.content)))
            return var
        return None


class GetInventory:
    @staticmethod
    def get(provider: str, **kwargs):
        saved_inventory: SpeasyIndex = index.get("proxy_inventories", provider, None)
        saved_inventory_dt: datetime = index.get("proxy_inventories_save_date", provider, datetime.utcfromtimestamp(0))
        if saved_inventory_dt + timedelta(days=inventories_cfg.cache_retention_days.get()) > datetime.utcnow():
            return saved_inventory

        url = proxy_cfg.url()
        kwargs['provider'] = provider
        kwargs['format'] = 'python_dict'
        kwargs['zstd_compression'] = zstd_compression
        headers = {}
        if saved_inventory is not None:
            headers["If-Modified-Since"] = parser.parse(saved_inventory.build_date).ctime()
        resp = http.get(f"{url}/get_inventory?", params=kwargs, headers=headers)
        log.debug(f"Asking {provider} inventory from proxy {resp.url}, {resp.request.headers}")
        if resp.status_code == 200:
            inventory = inventory_from_dict(pickle.loads(decompress(resp.content)))
            index.set("proxy_inventories", provider, inventory)
            index.set("proxy_inventories_save_date", provider, datetime.utcnow())
            return inventory
        if resp.status_code == 304:
            return saved_inventory
        return None


class Proxyfiable(object):
    def __init__(self, request, arg_builder):
        self.request = request
        self.arg_builder = arg_builder

    def __call__(self, func):
        @wraps(func)
        def wrapped(*args, **kwargs):
            disable_proxy = kwargs.pop("disable_proxy", False)
            if proxy_cfg.enabled() and not disable_proxy:
                try:
                    proxy_version = query_proxy_version()
                    if proxy_version is not None and proxy_version >= MINIMUM_REQUIRED_PROXY_VERSION:
                        return self.request.get(**self.arg_builder(**kwargs))
                    else:
                        log.warning(
                            f"You are using an incompatible proxy server {proxy_cfg.url()} which is {proxy_version} while minimun required version is {MINIMUM_REQUIRED_PROXY_VERSION}")
                except:  # lgtm [py/catch-base-exception]
                    log.error(f"Can't get data from proxy server {proxy_cfg.url()}")
            return func(*args, **kwargs)

        return wrapped



"""Base inventory tree management
"""
import xml.etree.ElementTree as Et

from ...core import fix_name
from ...core.inventory.indexes import (CatalogIndex, ComponentIndex,
                                       DatasetIndex, ParameterIndex,
                                       SpeasyIndex, TimetableIndex)
from ...inventories import flat_inventories


def to_xmlid(index_or_str) -> str:
    if type(index_or_str) is str:
        return index_or_str
    if type(index_or_str) is dict and "xmlid" in index_or_str:
        return index_or_str['xmlid']
    if hasattr(index_or_str, 'xmlid'):
        return index_or_str.xmlid
    else:
        raise TypeError(f"given parameter {index_or_str} of type {type(index_or_str)} is not a compatible index")


def to_parameter_index(index_or_str) -> ParameterIndex:
    if type(index_or_str) is str:
        if index_or_str in flat_inventories.amda.parameters:
            return flat_inventories.amda.parameters[index_or_str]
        else:
            raise ValueError(f"Unknown parameter: {index_or_str}")

    if type(index_or_str) is ParameterIndex:
        return index_or_str
    else:
        raise TypeError(f"given parameter {index_or_str} of type {type(index_or_str)} is not a compatible index")


def to_dataset_index(index_or_str) -> DatasetIndex:
    if type(index_or_str) is str:
        if index_or_str in flat_inventories.amda.datasets:
            return flat_inventories.amda.datasets[index_or_str]
        else:
            raise ValueError(f"Unknown dataset: {index_or_str}")

    if type(index_or_str) is DatasetIndex:
        return index_or_str
    else:
        raise TypeError(f"given dataset {index_or_str} of type {type(index_or_str)} is not a compatible index")


class AmdaXMLParser:

    @staticmethod
    def fix_names(**kwargs):
        return {fix_name(key): value for key, value in kwargs.items()}

    @staticmethod
    def fix_xmlid(**kwargs):
        def clean(key):
            if '}id' in key and key.startswith('{'):
                return 'xmlid'
            return key

        return {clean(key): value for key, value in kwargs.items()}

    @staticmethod
    def index_ctor_args(node, name_key='xmlid', is_public='True'):
        meta = AmdaXMLParser.fix_names(**AmdaXMLParser.fix_xmlid(**node.attrib))
        meta['is_public'] = is_public
        if 'dataStart' in meta:
            meta['start_date'] = meta.pop('dataStart')
        if 'dataStop' in meta:
            meta['stop_date'] = meta.pop('dataStop')
        uid = meta.get('xmlid', node.tag)
        return {"name": meta.get('name', node.tag), "provider": "amda", 'uid': uid, "meta": meta}

    @staticmethod
    def make_any_node(parent, node, ctor, name_key='xmlid', is_public='True'):
        new = ctor(**AmdaXMLParser.index_ctor_args(node, is_public=str(is_public)))
        name = fix_name(new.__dict__.get(name_key, node.tag))
        parent.__dict__[name] = new
        return new

    @staticmethod
    def make_instrument_node(parent, node, is_public=True):
        return AmdaXMLParser.make_any_node(parent, node, SpeasyIndex, name_key='name', is_public=is_public)

    @staticmethod
    def make_dataset_node(parent, node, is_public=True):
        ds = AmdaXMLParser.make_any_node(parent, node, DatasetIndex, is_public=is_public)
        return ds

    @staticmethod
    def make_parameter_node(parent, node, is_public=True):
        param = AmdaXMLParser.make_any_node(parent, node, ParameterIndex, is_public=is_public)
        if type(parent) is DatasetIndex:
            param.start_date = parent.start_date
            param.stop_date = parent.stop_date
            param.dataset = parent.spz_uid()
        return param

    @staticmethod
    def make_user_parameter_node(parent, node, is_public=True):
        # It seems that AMDA prevents users from using incompatible names here
        param = AmdaXMLParser.make_any_node(parent, node, ParameterIndex, name_key='name', is_public=is_public)
        return param

    @staticmethod
    def make_component_node(parent, node, is_public=True):
        component = AmdaXMLParser.make_any_node(parent, node, ComponentIndex, is_public=is_public)
        if type(parent) is ParameterIndex:
            component.start_date = parent.start_date
            component.stop_date = parent.stop_date
            component.dataset = parent.dataset
            component.parameter = parent.spz_uid()
        return component

    @staticmethod
    def make_timetable_node(parent, node, is_public=True):
        tt = AmdaXMLParser.make_any_node(parent, node, TimetableIndex, name_key='name', is_public=is_public)
        return tt

    @staticmethod
    def make_catalogue_node(parent, node, is_public=True):
        cat = AmdaXMLParser.make_any_node(parent, node, CatalogIndex, name_key='name', is_public=is_public)
        return cat

    @staticmethod
    def make_path_node(parent, node, is_public=True):
        return AmdaXMLParser.make_any_node(parent, node, SpeasyIndex, name_key='name', is_public=is_public)

    @staticmethod
    def parse(xml, is_public=True):
        handlers = {
            'instrument': AmdaXMLParser.make_instrument_node,
            'dataset': AmdaXMLParser.make_dataset_node,
            'parameter': AmdaXMLParser.make_parameter_node,
            'component': AmdaXMLParser.make_component_node,
            'timeTable': AmdaXMLParser.make_timetable_node,
            'timetab': AmdaXMLParser.make_timetable_node,
            'catalog': AmdaXMLParser.make_catalogue_node,
            'param': AmdaXMLParser.make_user_parameter_node,
        }

        def _recursive_parser(parent, node, is_public):
            new = handlers.get(node.tag, AmdaXMLParser.make_path_node)(parent, node, is_public)
            for subnode in node:
                _recursive_parser(new, subnode, is_public)

        root = SpeasyIndex("root", "amda", "amda_root_node")
        if xml is not None:
            tree = Et.fromstring(xml)
            _recursive_parser(root, tree, is_public=is_public)

        return root



from dataclasses import dataclass
from .mpl_backend import Plot as MplPlot
from ..core.data_containers import DataContainer, VariableAxis, VariableTimeAxis
from typing import List, Dict
import numpy as np
from enum import Enum
from copy import copy

__backends__ = {
    "matplotlib": MplPlot,
    None: MplPlot
}


class PlotType(Enum):
    LINE = 0
    SPECTRO = 1


@dataclass
class Plot:
    axes: List[VariableAxis or VariableTimeAxis]
    values: DataContainer
    columns_names: List[str]

    def _set_backend(self, name=None):
        if not hasattr(self, "_backend") or name != self._backend_name or self._backend_name is None:
            self._backend_name = name or "matplotlib"
            self._backend = __backends__[self._backend_name]()

    def _get_backend(self, name=None):
        self._set_backend(name)
        return self._backend

    def _with_backend(self, backend):
        new = copy(self)
        new._set_backend(backend)
        return new

    def _infer_plot_type(self):
        if self.values.meta.get("DISPLAY_TYPE", "") == "spectrogram":
            return PlotType.SPECTRO
        return PlotType.LINE

    def line(self, *args, backend=None, **kwargs):
        units = self.values.unit
        yaxis_label = self.values.name
        return self._get_backend(backend).line(x=self.axes[0].values, y=self.values.values, labels=self.columns_names,
                                               units=units,
                                               xaxis_label=self.axes[0].name,
                                               yaxis_label=yaxis_label, *args,
                                               **kwargs)

    def colormap(self, *args, logy=True, logz=True, backend=None, **kwargs):
        yaxis_units = self.axes[1].unit
        yaxis_label = self.axes[1].name
        zaxis_units = self.values.unit
        zaxis_label = self.values.name
        return self._get_backend(backend).colormap(x=self.axes[0].values, y=self.axes[1].values.T,
                                                   z=self.values.values.T,
                                                   xaxis_label=self.axes[0].name,
                                                   yaxis_label=yaxis_label,
                                                   yaxis_units=yaxis_units,
                                                   zaxis_label=zaxis_label,
                                                   zaxis_units=zaxis_units,
                                                   *args, **kwargs)

    def __call__(self, *args, backend=None, **kwargs):
        if self._infer_plot_type() == PlotType.SPECTRO:
            return self.colormap(backend=backend, *args, **kwargs)
        return self.line(backend=backend, *args, **kwargs)

    def __getitem__(self, item):
        assert type(item) is str
        return self._with_backend(item)



# -*- coding: utf-8 -*-

"""cda package for Space Physics WebServices Client."""

__author__ = """Alexis Jeandet"""
__email__ = 'alexis.jeandet@member.fsf.org'
__version__ = '0.1.0'

from typing import Optional, Dict
from datetime import datetime, timedelta
from speasy.core.cache import Cacheable, CacheCall, CACHE_ALLOWED_KWARGS
from speasy.products.variable import SpeasyVariable, VariableTimeAxis, DataContainer
from ...core import http, AllowedKwargs, deprecation
from speasy.core.proxy import Proxyfiable, GetProduct, PROXY_ALLOWED_KWARGS
from speasy.core.inventory.indexes import ParameterIndex, SpeasyIndex
from speasy.core.dataprovider import DataProvider, ParameterRangeCheck, GET_DATA_ALLOWED_KWARGS
from speasy.core.datetime_range import DateTimeRange
from speasy.core.requests_scheduling import SplitLargeRequests
import numpy as np
from astropy import units

import logging

log = logging.getLogger(__name__)


def _make_datetime(dt: str) -> np.datetime64:
    '''
        Hack to support python 3.6, once 3.6 support removed then go back to:
        datetime.strptime(v[1], '%Y-%m-%dT%H:%M:%S.%f%z').timestamp()
    '''
    return np.datetime64(dt[:-6], 'ns')


def _variable(orbit: dict) -> Optional[SpeasyVariable]:
    data = orbit['Result']['Data'][1][0]['Coordinates'][1][0]
    keys = list(data.keys())
    keys.remove('CoordinateSystem')
    values = np.array([data['X'][1], data['Y'][1], data['Z'][1]]).transpose()
    # this is damn slow!
    time_axis = np.array([_make_datetime(v[1]) for v in
                          orbit['Result']['Data'][1][0]['Time'][1]])
    return SpeasyVariable(
        axes=[VariableTimeAxis(values=time_axis)],
        values=DataContainer(values, meta={'CoordinateSystem': data['CoordinateSystem'], 'UNITS': 'km'}),
        columns=['X', 'Y', 'Z']
    )


def _is_valid(orbit: dict):
    return orbit['Result']['StatusCode'] == 'SUCCESS' and orbit['Result']['StatusSubCode'] == 'SUCCESS'


def _make_cache_entry_name(prefix: str, product: str, start_time: str, **kwargs):
    return f"{prefix}/{product}/{kwargs.get('coordinate_system', 'gse')}/{start_time}"


def get_parameter_args(start_time: datetime, stop_time: datetime, product: str, **kwargs):
    return {'path': f"sscweb/{product}", 'start_time': f'{start_time.isoformat()}',
            'stop_time': f'{stop_time.isoformat()}', 'coordinate_system': kwargs.get('coordinate_system', 'gse')}


def make_index(meta: Dict):
    name = meta.pop('Name')
    meta['start_date'] = meta.pop('StartTime')[1]
    meta['stop_date'] = meta.pop('EndTime')[1]
    node = ParameterIndex(name=name, provider="ssc", uid=meta['Id'], meta=meta)
    return node


class SSC_Webservice(DataProvider):
    def __init__(self):
        self.__url = "https://sscweb.gsfc.nasa.gov/WS/sscr/2"
        DataProvider.__init__(self, provider_name='ssc', provider_alt_names=['sscweb'])

    def build_inventory(self, root: SpeasyIndex):
        inv = list(map(make_index, self.get_observatories()))
        root.Trajectories = SpeasyIndex(name='Trajectories', provider='ssc', uid='Trajectories',
                                        meta={item.Id: item for item in inv})
        return root

    @CacheCall(cache_retention=7 * 24 * 60 * 60, is_pure=True)
    def get_observatories(self):
        res = http.get(f"{self.__url}/observatories", headers={"Accept": "application/json"})
        if not res.ok:
            return None
        return res.json()['Observatory'][1]

    def version(self, product):
        return 2

    def parameter_range(self, parameter_id: str or ParameterIndex) -> Optional[DateTimeRange]:
        """Get product time range.

        Parameters
        ----------
        parameter_id: str or ParameterIndex
            parameter id

        Returns
        -------
        Optional[DateTimeRange]
            Data time range

        Examples
        --------

        >>> import speasy as spz
        >>> spz.ssc.parameter_range("solarorbiter")
        <DateTimeRange: 2020-02-10T04:56:30+00:00 -> ...>

        """
        return self._parameter_range(parameter_id)

    def get_data(self, product: str, start_time: datetime, stop_time: datetime, coordinate_system: str = 'gse',
                 debug=False, **kwargs) -> Optional[SpeasyVariable]:
        var = self._get_orbit(product=product, start_time=start_time, stop_time=stop_time,
                              coordinate_system=coordinate_system, debug=debug, **kwargs)
        return var

    @AllowedKwargs(
        PROXY_ALLOWED_KWARGS + CACHE_ALLOWED_KWARGS + GET_DATA_ALLOWED_KWARGS + ['coordinate_system',
                                                                                 'debug'])
    @ParameterRangeCheck()
    @Cacheable(prefix="ssc_orbits", fragment_hours=lambda x: 24, version=version, entry_name=_make_cache_entry_name)
    @SplitLargeRequests(threshold=lambda: timedelta(days=60))
    @Proxyfiable(GetProduct, get_parameter_args)
    def _get_orbit(self, product: str, start_time: datetime, stop_time: datetime, coordinate_system: str = 'gse',
                   debug=False, extra_http_headers: Dict or None = None) -> Optional[SpeasyVariable]:
        if stop_time - start_time < timedelta(days=1):
            stop_time += timedelta(days=1)
        url = f"{self.__url}/locations/{product}/{start_time.strftime('%Y%m%dT%H%M%SZ')},{stop_time.strftime('%Y%m%dT%H%M%SZ')}/{coordinate_system.lower()}/"
        if debug:
            print(url)
        headers = {"Accept": "application/json"}
        if extra_http_headers is not None:
            headers.update(extra_http_headers)
        res = http.get(url, headers=headers)
        orbit = res.json()
        if res.ok and _is_valid(orbit):
            return _variable(orbit)[start_time:stop_time]
        return None



from datetime import datetime
from typing import Iterable, List, Optional, Tuple, Union, overload
import sys
import numpy as np

from ..inventory.indexes import (CatalogIndex, ComponentIndex,
                                 DatasetIndex, ParameterIndex,
                                 SpeasyIndex, TimetableIndex)

from ...products import *
from ...webservices import (AMDA_Webservice, CDA_Webservice, CSA_Webservice,
                            SSC_Webservice)
from .. import is_collection, progress_bar
from ..datetime_range import DateTimeRange
from ...config import core as core_cfg

TimeT = Union[str, datetime, float, np.datetime64]
TimeRangeT = Union[DateTimeRange, Tuple[TimeT, TimeT]]
TimeSerieIndexT = Union[ParameterIndex, ComponentIndex]
TimeRangeCollectionT = Union[TimetableIndex, CatalogIndex, Iterable[Iterable[Union[TimeT]]]]

PROVIDERS = {}
amda = None
csa = None
cda = None
ssc = None

if 'amda' not in core_cfg.disabled_providers():
    amda = AMDA_Webservice()
    sys.modules[__name__].amda = amda
    PROVIDERS['amda'] = amda

if 'csa' not in core_cfg.disabled_providers():
    csa = CSA_Webservice()
    sys.modules[__name__].csa = csa
    PROVIDERS['csa'] = csa

if not core_cfg.disabled_providers().intersection({'cda', 'cdaweb'}):
    cda = CDA_Webservice()
    sys.modules[__name__].cda = cda
    PROVIDERS['cda'] = cda
    PROVIDERS['cdaweb'] = cda

if not core_cfg.disabled_providers().intersection({'ssc', 'sscweb'}):
    ssc = SSC_Webservice()
    sys.modules[__name__].ssc = ssc
    PROVIDERS['ssc'] = ssc
    PROVIDERS['sscweb'] = ssc


def list_providers() -> List[str]:
    return list(PROVIDERS.keys())


@overload
def get_data(product: CatalogIndex, **kwargs) -> Catalog:
    ...


@overload
def get_data(product: TimetableIndex, **kwargs) -> TimeTable:
    ...


@overload
def get_data(product: str, **kwargs) -> TimeTable or Catalog:
    ...


@overload
def get_data(product: DatasetIndex, start_time: TimeT, stop_time: TimeT, **kwargs) -> Dataset or None:
    ...


@overload
def get_data(product: DatasetIndex, time_range: TimeRangeT, **kwargs) -> Dataset or None:
    ...


@overload
def get_data(product: DatasetIndex, time_range: Iterable[TimeRangeT], **kwargs) -> List[Optional[Dataset]] or None:
    ...


@overload
def get_data(product: TimeSerieIndexT, start_time: TimeT, stop_time: TimeT, **kwargs) -> SpeasyVariable or None:
    ...


@overload
def get_data(product: TimeSerieIndexT, time_range: TimeRangeT, **kwargs) -> SpeasyVariable or None:
    ...


@overload
def get_data(product: TimeSerieIndexT, time_range: Iterable[TimeRangeT], **kwargs) -> List[Optional[
    SpeasyVariable]] or None:
    ...


@overload
def get_data(product: TimeSerieIndexT, time_range: TimeRangeCollectionT, **kwargs) -> List[Optional[
    SpeasyVariable]] or None:
    ...


@overload
def get_data(product: Iterable[TimeSerieIndexT], start_time: TimeT, stop_time: TimeT,
             **kwargs) -> List[Optional[SpeasyVariable]] or None:
    ...


@overload
def get_data(product: Iterable[TimeSerieIndexT], time_range: TimeRangeCollectionT, **kwargs) -> List[List[
    Optional[SpeasyVariable]]] or None:
    ...


def _could_be_datetime(value):
    return type(value) in (str, datetime, np.datetime64, float)


def provider_and_product(path_or_product: str or SpeasyIndex) -> (str, str):
    """Breaks given product in two parts: provider and product UID

    Parameters
    ----------
    path_or_product : str or SpeasyIndex
        The product as SpeasyIndex or path as string you want to split
    Returns
    -------
    (str, str)
        the provider UID and the product UID
    """
    if isinstance(path_or_product, SpeasyIndex):
        return path_or_product.spz_provider().lower(), path_or_product.spz_uid()
    elif type(path_or_product) is str:
        if '/' in path_or_product:
            provider_uid, product_uid = path_or_product.split('/', 1)
            return provider_uid, product_uid
        else:
            raise ValueError(f"Given string does not look like a path {path_or_product}")
    raise TypeError(
        f"Wrong type for {path_or_product}, expecting a string or a SpeasyIndex, got {type(path_or_product)}")


def _scalar_get_data(index, *args, **kwargs):
    provider_uid, product_uid = provider_and_product(index)
    if provider_uid in PROVIDERS:
        return PROVIDERS[provider_uid].get_data(product_uid, *args, **kwargs)
    raise ValueError(f"Can't find a provider for {index}")


def _get_catalog_or_timetable(index, **kwargs):
    return _scalar_get_data(index, **kwargs)


def _get_timeserie1(index, dtrange, **kwargs):
    return _scalar_get_data(index, dtrange[0], dtrange[1], **kwargs)


def _get_timeserie2(index, start, stop, **kwargs):
    return _scalar_get_data(index, start, stop, **kwargs)


def _compile_args(*args, **kwargs):
    if len(args) == 0:
        if 'product' in kwargs:
            args = [kwargs.pop('product')]
    if len(args) == 1:
        if 'start_time' in kwargs and 'stop_time' in kwargs:
            args += [kwargs.pop('start_time'), kwargs.pop('stop_time')]
        elif 'time_range' in kwargs:
            args += [kwargs.pop('time_range')]
    return args, kwargs


def _is_dtrange(value):
    return type(value) is DateTimeRange or (
        hasattr(value, '__len__') and len(value) == 2 and _could_be_datetime(value[0]) and _could_be_datetime(value[1]))


def get_data(*args, **kwargs) -> MaybeAnyProduct:
    """Retrieve requested product(s).
    Speasy gives access to two kind of products, time-dependent products such as physical measurements or trajectories
    and products such as timetables or event catalogs. So depending on which product you want to retrieve :func:`speasy.get_data`
    accepts different sets of arguments:

    - For time-independent products:

    .. highlight:: python
    .. code-block:: python

        get_data(product_or_products, **kwargs)

    - For time-dependent products:

    .. highlight:: python
    .. code-block:: python

        get_data(product_or_products, start_time, stop_time, **kwargs)
        get_data(product_or_products, datetime_range_or_datetime_range, **kwargs)

    Since get_data accepts both at the same time a list of products and a list of ranges, it will always iterate first
    on products then on datetime ranges. In other words, all products will be retrieved for all given datetime ranges.

    Parameters
    ----------
    args :
        Either a time independent or a list of time independent indexes or any combination of time dependent or list of
        time dependent indexes plus a datetime range or a list of datetime ranges. See examples below for more details.
    kwargs :
        For webservice specific keyword arguments check :doc:`/user/webservices`.

        - disable_proxy: bool
            ignore proxy configuration and always bypass proxy server when True (default: False).
        - disable_cache: bool
            ignore cache content when True (default: False).
        - progress: bool
            show progress bar when True (default: False).

    Returns
    -------
        requested product(s) according to given parameters, either a single product or a collection of products.

    Examples
    --------

    - A simple parameter request on a single tile range:

    >>> import speasy as spz
    >>> spz.get_data("amda/imf_gsm", "2016-10-10", "2016-10-11")
    <speasy.products.variable.SpeasyVariable object at ...>


    - Same with a catalog using Speasy dynamic inventory:

    >>> import speasy as spz
    >>> amda_catalogs = spz.inventories.tree.amda.Catalogs
    >>> spz.get_data(amda_catalogs.SharedCatalogs.EARTH.model_regions_plasmas_cluster_2005)
    <Catalog: model_regions_plasmas_cluster_2005>

    - You can also request a collection of catalogs:

    >>> import speasy as spz
    >>> amda_catalogs = spz.inventories.tree.amda.Catalogs
    >>> spz.get_data([amda_catalogs.SharedCatalogs.EARTH.model_regions_plasmas_cluster_2005,
    ...               amda_catalogs.SharedCatalogs.EARTH.model_regions_plasmas_mms_2019])
    [<Catalog: model_regions_plasmas_cluster_2005>, <Catalog: model_regions_plasmas_mms_2019>]

    - You can also request a parameter for several intervals:

    >>> import speasy as spz
    >>> spz.get_data("amda/imf_gsm", [["2016-10-10", "2016-10-11"],
    ...                               ["2017-10-10", "2017-10-11"]])
    [<speasy.products.variable.SpeasyVariable object at ...>, <speasy.products.variable.SpeasyVariable object at ...>]

    - Several products on a single interval:

    >>> import speasy as spz
    >>> spz.get_data(["amda/imf_gsm", spz.inventories.tree.ssc.Trajectories.wind],
    ...              "2016-10-10", "2016-10-11")
    [<speasy.products.variable.SpeasyVariable object at ...>, <speasy.products.variable.SpeasyVariable object at ...>]

    - Several products for several time ranges:

    >>> import speasy as spz
    >>> data= spz.get_data(["amda/imf_gsm",
    ...                  spz.inventories.tree.ssc.Trajectories.wind],
    ...              [["2016-10-10", "2016-10-11"],
    ...               ["2017-10-10", "2017-10-11"]])
    >>> len(data), len(data[0])
    (2, 2)

    - A catalog or a timetable can also be used as time ranges collection to download a product:

    >>> import speasy as spz
    >>> amda_shared_tt = spz.inventories.tree.amda.TimeTables.SharedTimeTables
    >>> mex_inventory = spz.inventories.tree.amda.Parameters.MEX
    >>> mgs_inventory = spz.inventories.tree.amda.Parameters.MGS
    >>> conj_mex_mgs = spz.get_data(amda_shared_tt.MARS.conjonctions_mex_mgs_2004_0)
    >>> data = spz.get_data(
    ...                     [mex_inventory.ELS.mex_els_all.mex_els_spec,
    ...                      mgs_inventory.MAG.mgs_mag_mso.b_mgs_mso],
    ...                     conj_mex_mgs)
    >>> len(data), len(data[0])
    (2, 28)

    - Can even pass a CatalogIndex or a TimeTableIndex directly:

    >>> import speasy as spz
    >>> amda_shared_tt = spz.inventories.tree.amda.TimeTables.SharedTimeTables
    >>> mex_inventory = spz.inventories.tree.amda.Parameters.MEX
    >>> mgs_inventory = spz.inventories.tree.amda.Parameters.MGS
    >>> data = spz.get_data(
    ...                     [mex_inventory.ELS.mex_els_all.mex_els_spec,
    ...                      mgs_inventory.MAG.mgs_mag_mso.b_mgs_mso],
    ...                     amda_shared_tt.MARS.conjonctions_mex_mgs_2004_0)
    >>> len(data), len(data[0])
    (2, 28)

    """
    args, kwargs = _compile_args(*args, **kwargs)
    if len(args) == 0:
        raise ValueError("You must at least provide a product to retrieve")

    product = args[0]
    if is_collection(product) and not isinstance(product, SpeasyIndex):
        return list(map(lambda p: get_data(p, *args[1:], **kwargs), progress_bar(leave=True, **kwargs)(product)))

    if len(args) == 1:
        return _get_catalog_or_timetable(*args, **kwargs)
    if len(args) == 2:
        t_range = args[1]
        if _is_dtrange(t_range):
            return _get_timeserie1(*args, **kwargs)
        if is_collection(t_range):
            return list(
                map(lambda r: get_data(product, r, *args[2:], **kwargs),
                    progress_bar(leave=False, **kwargs)(t_range)))
        return get_data(product, get_data(t_range), *args[2:], **kwargs)
    if len(args) == 3:
        return _get_timeserie2(*args, **kwargs)



import logging
from datetime import datetime
from functools import wraps
from typing import Callable, List, Optional
from threading import Lock

from speasy.core.datetime_range import DateTimeRange
from speasy.core.inventory import ProviderInventory
from speasy.core.inventory.indexes import (DatasetIndex, ParameterIndex,
                                           SpeasyIndex, inventory_has_changed)
from speasy.core.proxy import GetInventory, Proxyfiable
from speasy.inventories import flat_inventories, tree

log = logging.getLogger(__name__)
GET_DATA_ALLOWED_KWARGS = ['product', 'start_time', 'stop_time', 'extra_http_headers', 'progress']
PROVIDERS = {}


class ParameterRangeCheck(object):
    def __init__(self):
        pass

    def __call__(self, get_data: Callable):
        @wraps(get_data)
        def wrapped(wrapped_self, product, start_time, stop_time, **kwargs):
            p_range = wrapped_self.parameter_range(product)
            if not p_range.intersect(DateTimeRange(start_time, stop_time)):
                log.warning(f"You are requesting {product} outside of its definition range {p_range}")
                return None
            return get_data(wrapped_self, product=product, start_time=start_time, stop_time=stop_time, **kwargs)

        return wrapped


def _get_inventory_args(provider_name, **kwargs):
    return {'provider': f"{provider_name}"}


class DataProvider:
    def __init__(self, provider_name: str, provider_alt_names: List or None = None):
        self.provider_name = provider_name
        self.provider_alt_names = provider_alt_names or []
        self.flat_inventory = ProviderInventory()
        flat_inventories.__dict__[provider_name] = self.flat_inventory
        for alt_name in self.provider_alt_names:
            flat_inventories.__dict__[alt_name] = self.flat_inventory
        self.update_inventory()
        PROVIDERS[provider_name] = self

    @Proxyfiable(request=GetInventory, arg_builder=_get_inventory_args)
    def _inventory(self, provider_name) -> SpeasyIndex:
        return self.build_inventory(SpeasyIndex(provider=provider_name, name=provider_name, uid=provider_name,
                                                meta={'build_date': datetime.utcnow().isoformat()}))

    def _update_private_inventory(self, root: SpeasyIndex):
        if hasattr(self, 'build_private_inventory'):
            return self.build_private_inventory(root)

    def update_inventory(self):
        lock = Lock()
        with lock:
            new_inventory = self._inventory(provider_name=self.provider_name)
            if inventory_has_changed(tree.__dict__.get(self.provider_name, SpeasyIndex("", "", "")), new_inventory):
                if self.provider_name in tree.__dict__:
                    tree.__dict__[self.provider_name].clear()
                tree.__dict__[self.provider_name] = new_inventory
            self._update_private_inventory(tree.__dict__[self.provider_name])
            self.flat_inventory.clear()
            self.flat_inventory.update(tree.__dict__[self.provider_name])

    def _to_dataset_index(self, index_or_str) -> DatasetIndex:
        if type(index_or_str) is str:
            if index_or_str in self.flat_inventory.datasets:
                return self.flat_inventory.datasets[index_or_str]
            else:
                raise ValueError(f"Unknown dataset: {index_or_str}")

        if type(index_or_str) is DatasetIndex:
            return index_or_str
        else:
            raise TypeError(f"given dataset {index_or_str} of type {type(index_or_str)} is not a compatible index")

    def _to_parameter_index(self, index_or_str) -> ParameterIndex:
        if type(index_or_str) is str:
            if index_or_str in self.flat_inventory.parameters:
                return self.flat_inventory.parameters[index_or_str]
            else:
                raise ValueError(f"Unknown parameter: {index_or_str}")

        if type(index_or_str) is ParameterIndex:
            return index_or_str
        else:
            raise TypeError(f"given parameter {index_or_str} of type {type(index_or_str)} is not a compatible index")

    def _parameter_range(self, parameter_id: str or ParameterIndex) -> Optional[DateTimeRange]:
        parameter = self._to_parameter_index(parameter_id)
        return DateTimeRange(
            parameter.start_date,
            parameter.stop_date
        )

    def _dataset_range(self, dataset_id: str or DatasetIndex) -> Optional[DateTimeRange]:
        ds = self._to_dataset_index(dataset_id)
        return DateTimeRange(
            ds.start_date,
            ds.stop_date
        )



import logging
from datetime import datetime, timedelta
from types import SimpleNamespace
from typing import Dict, Optional

# General modules
from ...config import amda as amda_cfg
from ...core.cache import CacheCall
from ...core.inventory.indexes import SpeasyIndex
from ...inventories import flat_inventories
from ...products.variable import SpeasyVariable, merge
from . import rest_client
from .exceptions import MissingCredentials
from .inventory import AmdaXMLParser
from .rest_client import auth_args
from .utils import load_catalog, load_csv, load_timetable

log = logging.getLogger(__name__)


def credential_are_valid():
    login = amda_cfg.username()
    password = amda_cfg.password()
    return login != "" and password != ""


def _get_credentials():
    if credential_are_valid():
        return amda_cfg.username(), amda_cfg.password()
    else:
        raise MissingCredentials()


def is_public(node):
    return node.__dict__.get('is_public', 'True') == 'True'


def is_private(node):
    return not is_public(node)


class AmdaImpl:
    def __init__(self, server_url: str = amda_cfg.entry_point()):
        self.server_url = server_url

    def _update_private_lists(self, TimeTables: SpeasyIndex, Catalogs: SpeasyIndex, root: SpeasyIndex):
        if credential_are_valid():
            username, password = _get_credentials()
            user_tt = AmdaXMLParser.parse(
                rest_client.get_user_timetables_xml_tree(username=username, password=password,
                                                         server_url=self.server_url),
                is_public=False)
            TimeTables.MyTimeTables = SpeasyIndex(name='MyTimeTables', provider='amda', uid='MyTimeTables',
                                                  meta=user_tt.timetabList.__dict__)

            user_cat = AmdaXMLParser.parse(
                rest_client.get_user_catalogs_xml_tree(username=username, password=password,
                                                       server_url=self.server_url),
                is_public=False)
            Catalogs.MyCatalogs = SpeasyIndex(name='MyCatalogs', provider='amda', uid='MyCatalogs',
                                              meta=user_cat.catalogList.__dict__)

            user_param = AmdaXMLParser.parse(
                rest_client.get_user_parameters_xml_tree(username=username, password=password,
                                                         server_url=self.server_url),
                is_public=False)
            root.DerivedParameters = SpeasyIndex(name='DerivedParameters', provider='amda', uid='DerivedParameters',
                                                 meta=user_param.ws.paramList.__dict__)

    def _update_lists(self, TimeTables: SpeasyIndex, Catalogs: SpeasyIndex, root: SpeasyIndex):

        public_tt = AmdaXMLParser.parse(rest_client.get_timetables_xml_tree(server_url=self.server_url))
        TimeTables.SharedTimeTables = SpeasyIndex(name='SharedTimeTables', provider='amda', uid='SharedTimeTables',
                                                  meta=public_tt.timeTableList.__dict__)
        public_cat = AmdaXMLParser.parse(rest_client.get_catalogs_xml_tree(server_url=self.server_url))
        Catalogs.SharedCatalogs = SpeasyIndex(name='SharedCatalogs', provider='amda', uid='SharedCatalogs',
                                              meta=public_cat.catalogList.__dict__)

    def build_inventory(self, root: SpeasyIndex):
        data = AmdaXMLParser.parse(rest_client.get_obs_data_tree(server_url=self.server_url))
        root.Parameters = SpeasyIndex(name='Parameters', provider='amda', uid='Parameters',
                                      meta=data.dataRoot.AMDA.__dict__)

        root.TimeTables = SpeasyIndex(name='TimeTables', provider='amda', uid='TimeTables')
        root.Catalogs = SpeasyIndex(name='Catalogs', provider='amda', uid='Catalogs')
        root.DerivedParameters = SpeasyIndex(name='DerivedParameters', provider='amda', uid='DerivedParameters')
        self._update_lists(TimeTables=root.TimeTables, Catalogs=root.Catalogs, root=root)
        return root

    def build_private_inventory(self, root: SpeasyIndex):
        self._update_private_lists(TimeTables=root.TimeTables, Catalogs=root.Catalogs, root=root)
        return root

    def dl_parameter_chunk(self, start_time: datetime, stop_time: datetime, parameter_id: str,
                           extra_http_headers: Dict or None = None, **kwargs) -> Optional[SpeasyVariable]:
        url = rest_client.get_parameter(server_url=self.server_url, startTime=start_time.timestamp(),
                                        stopTime=stop_time.timestamp(), parameterID=parameter_id, timeFormat='UNIXTIME',
                                        extra_http_headers=extra_http_headers, **kwargs)
        # check status until done
        if url is not None:
            var = load_csv(url)
            if len(var):
                log.debug(
                    f'Loaded var: data shape = {var.values.shape}, data start time = {var.time[0]}, \
                            data stop time = {var.time[-1]}')
            else:
                log.debug('Loaded var: Empty var')
            return var
        return None

    def dl_parameter(self, start_time: datetime, stop_time: datetime, parameter_id: str,
                     extra_http_headers: Dict or None = None, **kwargs) -> Optional[SpeasyVariable]:
        dt = timedelta(days=amda_cfg.max_chunk_size_days())

        if stop_time - start_time > dt:
            var = None
            curr_t = start_time
            while curr_t < stop_time:
                if curr_t + dt < stop_time:
                    var = merge([var, self.dl_parameter_chunk(curr_t, curr_t + dt, parameter_id,
                                                              extra_http_headers=extra_http_headers, **kwargs)])
                else:
                    var = merge([var, self.dl_parameter_chunk(curr_t, stop_time, parameter_id,
                                                              extra_http_headers=extra_http_headers, **kwargs)])
                curr_t += dt
            return var
        else:
            return self.dl_parameter_chunk(start_time, stop_time, parameter_id, extra_http_headers=extra_http_headers,
                                           **kwargs)

    def dl_user_parameter(self, start_time: datetime, stop_time: datetime, parameter_id: str,
                          **kwargs) -> Optional[SpeasyVariable]:
        username, password = _get_credentials()
        return self.dl_parameter(parameter_id=parameter_id, start_time=start_time, stop_time=stop_time,
                                 **auth_args(username=username, password=password), **kwargs)

    @CacheCall(cache_retention=amda_cfg.user_cache_retention())
    def dl_timetable(self, timetable_id: str, **kwargs):
        url = rest_client.get_timetable(ttID=timetable_id, server_url=self.server_url, **kwargs)
        if url is not None:
            timetable = load_timetable(filename=url)
            if timetable:
                timetable.meta.update(flat_inventories.amda.timetables.get(timetable_id, SimpleNamespace()).__dict__)
                log.debug(f'Loaded timetable: id = {timetable_id}')  # lgtm[py/clear-text-logging-sensitive-data]
            else:
                log.debug('Got None')
            return timetable
        return None

    def dl_user_timetable(self, timetable_id: str, **kwargs):
        username, password = _get_credentials()
        return self.dl_timetable(timetable_id, **auth_args(username=username, password=password), **kwargs)

    @CacheCall(cache_retention=amda_cfg.user_cache_retention())
    def dl_catalog(self, catalog_id: str, **kwargs):
        url = rest_client.get_catalog(catID=catalog_id, server_url=self.server_url, **kwargs)
        if url is not None:
            catalog = load_catalog(url)
            if catalog:
                log.debug(f'Loaded catalog: id = {catalog_id}')  # lgtm[py/clear-text-logging-sensitive-data]
                catalog.meta.update(flat_inventories.amda.catalogs.get(catalog_id, SimpleNamespace()).__dict__)
            else:
                log.debug('Got None')
            return catalog
        return None

    def dl_user_catalog(self, catalog_id: str, **kwargs):
        username, password = _get_credentials()
        return self.dl_catalog(catalog_id, **auth_args(username=username, password=password), **kwargs)



from copy import deepcopy
from typing import Dict, List, Optional

import astropy.table
import astropy.units
import numpy as np
import pandas as pds

from speasy.core.data_containers import (
    DataContainer,
    VariableAxis,
    VariableTimeAxis,
    _to_index,
)
from speasy.plotting import Plot

from .base_product import SpeasyProduct


class SpeasyVariable(SpeasyProduct):
    """SpeasyVariable object. Base class for storing variable data.

    Attributes
    ----------
    time: numpy.ndarray
        time vector (x-axis data)
    values: numpy.ndarray
        data
    meta: Optional[dict]
        metadata
    columns: Optional[List[str]]
        column names, might be empty for spectrograms or 3D+ data
    axes: List[np.ndarray]
        Collection composed of time axis plus eventual additional axes according to values' shape
    axes_labels: List[str]
        Axes names
    unit: str
        Values physical unit
    name: str
        SpeasyVariable name
    nbytes: int
        memory usage in bytes

    Methods
    -------
    view:
        Returns a view of the current variable within the desired :data:`index_range`
    to_dataframe:
        Converts the variable to a pandas.DataFrame object
    from_dataframe:
        Builds a SpeasyVariable from a pandas.DataFrame object
    to_astropy_table:
        Converts the variable to an astropy.table.Table
    unit_applied:
        Returns a copy where values are astropy.units.Quantity
    filter_columns:
        Returns a copy only containing selected columns
    replace_fillval_by_nan:
        Returns a SpeasyVaraible with NaN instead of fill value if fill value is set in meta data
    plot:
        Plot the data with matplotlib by default
    to_dictionary:
        Converts a SpeasyVariable to a Python dictionary, mostly used for serialization purposes
    copy:
        Returns a copy

    """

    __slots__ = ["__values_container", "__columns", "__axes"]

    def __init__(
        self,
        axes: List[VariableAxis or VariableTimeAxis],
        values: DataContainer,
        columns: Optional[List[str]] = None,
    ):
        super().__init__()
        if not isinstance(axes[0], VariableTimeAxis):
            raise TypeError(
                f"axes[0] must be a VariableTimeAxis instance, got {type(axes[0])}"
            )
        if axes[0].shape[0] != values.shape[0]:
            raise ValueError(
                f"Time and data must have the same length, got time:{len(axes[0])} and data:{len(values)}"
            )

        self.__columns = list(map(str.strip, columns or []))
        if len(values.values.shape) == 1:
            # to be consistent with pandas
            values.reshape((values.shape[0], 1))

        self.__values_container = values
        self.__axes = axes

    def view(self, index_range: slice) -> "SpeasyVariable":
        """Return view of the current variable within the desired :data:`index_range`.

        Parameters
        ----------
        index_range: slice
            index range

        Returns
        -------
        speasy.common.variable.SpeasyVariable
            view of the variable on the given range
        """
        return SpeasyVariable(
            axes=[
                axis[index_range] if axis.is_time_dependent else axis
                for axis in self.__axes
            ],
            values=self.__values_container[index_range],
            columns=self.columns,
        )

    def copy(self) -> "SpeasyVariable":
        """Makes a deep copy the variable

        Returns
        -------
        SpeasyVariable
            deep copy the variable
        """
        return SpeasyVariable(
            axes=deepcopy(self.__axes),
            values=deepcopy(self.__values_container),
            columns=deepcopy(self.columns),
        )

    def filter_columns(self, columns: List[str]) -> "SpeasyVariable":
        """Builds a SpeasyVariable with only selected columns

        Parameters
        ----------
        columns : List[str]
            list of column names to keep

        Returns
        -------
        SpeasyVariable
            a SpeasyVariable with only selected columns
        """
        indexes = list(map(lambda v: self.__columns.index(v), columns))
        return SpeasyVariable(
            axes=deepcopy(self.__axes),
            values=DataContainer(
                is_time_dependent=self.__values_container.is_time_dependent,
                name=self.__values_container.name,
                meta=deepcopy(self.__values_container.meta),
                values=self.__values_container.values[:, indexes],
            ),
            columns=columns,
        )

    def __eq__(self, other: "SpeasyVariable") -> bool:
        """Check if this variable equals another.

        Parameters
        ----------
        other: speasy.common.variable.SpeasyVariable
            another SpeasyVariable object to compare with

        Returns
        -------
        bool:
            True if all attributes are equal
        """
        return (
            type(other) is SpeasyVariable
            and self.__axes == other.__axes
            and self.__values_container == other.__values_container
        )

    def __len__(self):
        return len(self.__axes[0])

    def __getitem__(self, key):
        if isinstance(key, slice):
            return self.view(
                slice(_to_index(key.start, self.time),
                      _to_index(key.stop, self.time))
            )
        if type(key) in (list, tuple) and all(map(lambda v: type(v) is str, key)):
            return self.filter_columns(key)
        if type(key) is str and key in self.__columns:
            return self.filter_columns([key])
        raise ValueError(
            f"No idea how to slice SpeasyVariable with given value: {key}")

    def __setitem__(self, k, v: "SpeasyVariable"):
        assert type(v) is SpeasyVariable
        self.__values_container[k] = v.__values_container
        for axis, src_axis in zip(self.__axes, v.__axes):
            if axis.is_time_dependent:
                axis[k] = src_axis

    @property
    def name(self) -> str:
        """SpeasyVariable name

        Returns
        -------
        str
            SpeasyVariable name
        """
        return self.__values_container.name

    @property
    def values(self) -> np.array:
        """SpeasyVariable values

        Returns
        -------
        np.array
            SpeasyVariable values
        """
        return self.__values_container.values

    @property
    def time(self) -> np.array:
        """Time axis values, equivalent to var.axes[0].values

        Returns
        -------
        np.array
            time axis values as numpy array of datetime64[ns]
        """
        return self.__axes[0].values

    @property
    def meta(self) -> Dict:
        """SpeasyVariable meta-data

        Returns
        -------
        Dict
            SpeasyVariable meta-data
        """
        return self.__values_container.meta

    @property
    def axes(self) -> List[VariableTimeAxis or VariableAxis]:
        """SpeasyVariable axes, axis 0 is always a VariableTimeAxis, there should be the same number of axes than values dimensions

        Returns
        -------
        List[VariableTimeAxis or VariableAxis]
            list of variable axes
        """
        return self.__axes

    @property
    def axes_labels(self) -> List[str]:
        """Axes names respecting axes order

        Returns
        -------
        List[str]
            list of axes names
        """
        return [axis.name for axis in self.__axes]

    @property
    def columns(self) -> List[str]:
        """SpeasyVariable columns names when it makes sense

        Returns
        -------
        List[str]
            list of columns names
        """
        return self.__columns

    @property
    def unit(self) -> str:
        """SpeasyVariable unit if found in meta-data

        Returns
        -------
        str
            unit if found in meta-data
        """
        return self.__values_container.unit

    @property
    def nbytes(self) -> int:
        """SpeasyVariable's values and axes memory usage

        Returns
        -------
        int
            number of bytes used to store values and axes
        """
        return self.__values_container.nbytes + np.sum(
            list(map(lambda ax: ax.nbytes, self.__axes))
        )

    def unit_applied(self, unit: str or None = None, copy=True) -> "SpeasyVariable":
        """Returns a SpeasyVariable with given or automatically found unit applied to values

        Parameters
        ----------
        unit : str or None, optional
            Use given unit or gets one from variable metadata, by default None
        copy : bool, optional
            Preserves source variable and returns a modified copy if true, by default True

        Returns
        -------
        SpeasyVariable
            SpeasyVariable identic to source one with values converted to astropy.units.Quantity according to given or found unit

        See Also
        --------
        unit: returns variable unit if found in meta-data
        """
        if copy:
            axes = deepcopy(self.__axes)
            values = deepcopy(self.__values_container)
            columns = deepcopy(self.__columns)
        else:
            axes = self.__axes
            values = self.__values_container
            columns = self.__columns
        return SpeasyVariable(
            axes=axes, values=values.unit_applied(unit), columns=columns
        )

    def to_astropy_table(self) -> astropy.table.Table:
        """Convert the variable to an astropy.Table object.

        Parameters
        ----------
        datetime_index: bool
            boolean indicating that the index is datetime

        Returns
        -------
        astropy.Table:
            Variable converted to astropy.Table

        See Also
        --------
        from_dataframe: builds a SpeasyVariable from a pandas DataFrame
        to_dataframe: exports a SpeasyVariable to a pandas DataFrame
        """
        try:
            units = astropy.units.Unit(self.meta["UNITS"])
        except (ValueError, KeyError):
            units = None
        df = self.to_dataframe()
        umap = {c: units for c in df.columns}
        return astropy.table.Table.from_pandas(df, units=umap, index=True)

    def to_dataframe(self) -> pds.DataFrame:
        """Convert the variable to a pandas.DataFrame object.

        Returns
        -------
        pandas.DataFrame:
            Variable converted to Pandas DataFrame

        See Also
        --------
        from_dataframe: builds a SpeasyVariable from a pandas DataFrame
        to_astropy_table: exports a SpeasyVariable to an astropy.Table object
        """
        if len(self.__values_container.shape) != 2:
            raise ValueError(
                f"Cant' convert a SpeasyVariable with shape {self.__values_container.shape} to DataFrame, only 1D/2D variables are accepted"
            )
        return pds.DataFrame(
            index=self.time, data=self.values, columns=self.__columns, copy=True
        )

    @staticmethod
    def from_dataframe(df: pds.DataFrame) -> "SpeasyVariable":
        """Load from pandas.DataFrame object.

        Parameters
        ----------
        df: pandas.DataFrame
            Input DataFrame to convert

        Returns
        -------
        SpeasyVariable:
            Variable created from DataFrame

        See Also
        --------
        to_dataframe: exports a SpeasyVariable to a pandas DataFrame
        to_astropy_table: exports a SpeasyVariable to an astropy.Table object
        """
        if df.index.dtype == np.dtype("datetime64[ns]"):
            time = np.array(df.index)
        elif hasattr(df.index[0], "timestamp"):
            time = np.array(
                [np.datetime64(d.timestamp() * 1e9, "ns") for d in df.index]
            )
        else:
            raise ValueError(
                "Can't convert DataFrame index to datetime64[ns] array")
        return SpeasyVariable(
            axes=[VariableTimeAxis(values=time, meta={})],
            values=DataContainer(values=df.values, meta={}, name="Unknown"),
            columns=list(df.columns),
        )

    def to_dictionary(self, array_to_list=False) -> Dict[str, object]:
        """Converts SpeasyVariable to dictionary

        Parameters
        ----------
        array_to_list : bool, optional
            Converts numpy arrays to Python Lists when true, by default False

        Returns
        -------
        Dict[str, object]

        See Also
        --------
        from_dictionary: builds variable from dictionary
        """
        return {
            "axes": [
                axis.to_dictionary(array_to_list=array_to_list) for axis in self.__axes
            ],
            "values": self.__values_container.to_dictionary(
                array_to_list=array_to_list
            ),
            "columns": deepcopy(self.__columns),
        }

    @staticmethod
    def from_dictionary(dictionary: Dict[str, object] or None) -> "SpeasyVariable" or None:
        """Builds a SpeasyVariable from a well formed dictionary

        Returns
        -------
        SpeasyVariable or None

        See Also
        --------
        to_dictionary: exports SpeasyVariable to dictionary
        """
        if dictionary is not None:
            axes = dictionary["axes"]
            axes = [VariableTimeAxis.from_dictionary(axes[0])] + [
                VariableAxis.from_dictionary(axis) for axis in axes[1:]
            ]

            return SpeasyVariable(
                values=DataContainer.from_dictionary(dictionary["values"]),
                axes=axes,
                columns=dictionary.get("columns", None),
            )
        else:
            return None

    @property
    def plot(self, *args, **kwargs):
        """Plot the variable, tries to do its best to detect variable type and to populate plot labels

        """
        return Plot(
            values=self.__values_container, columns_names=self.columns, axes=self.axes
        )

    def replace_fillval_by_nan(self, inplace=False) -> "SpeasyVariable":
        """Replaces fill values by NaN, non float values are automatically converted to float.
        Fill value is taken from metadata field "FILLVAL"

        Parameters
        ----------
        inplace : bool, optional
            Modifies source variable when true else modifies and returns a copy, by default False

        Returns
        -------
        SpeasyVariable
            source variable or copy with fill values replaced by NaN
        """
        if inplace:
            res = self
        else:
            res = deepcopy(self)
        if "FILLVAL" in res.meta:
            res.__values_container.replace_val_by_nan(res.meta["FILLVAL"])
        return res

    @staticmethod
    def reserve_like(other: "SpeasyVariable", length: int = 0) -> "SpeasyVariable":
        """Create a SpeasyVariable of given length and with the same properties than given variable but unset values

        Parameters
        ----------
        other : SpeasyVariable
            variable used as reference for shape and meta-data
        length : int, optional
            output variable length, by default 0

        Returns
        -------
        SpeasyVariable
            a SpeasyVariable similar to given one of given length
        """
        axes = []
        for axis in other.__axes:
            if axis.is_time_dependent:
                new_axis = type(axis).reserve_like(axis, length)
                axes.append(new_axis)
            else:
                axes.append(deepcopy(axis))
        return SpeasyVariable(
            values=DataContainer.reserve_like(
                other.__values_container, length),
            axes=axes,
            columns=other.columns,
        )


def to_dictionary(var: SpeasyVariable, array_to_list=False) -> Dict[str, object]:
    return var.to_dictionary(array_to_list=array_to_list)


def from_dictionary(dictionary: Dict[str, object] or None) -> SpeasyVariable or None:
    return SpeasyVariable.from_dictionary(dictionary)


def from_dataframe(df: pds.DataFrame) -> SpeasyVariable:
    """Convert a dataframe to SpeasyVariable.

    See Also
    --------
    SpeasyVariable.from_dataframe
    """
    return SpeasyVariable.from_dataframe(df)


def to_dataframe(var: SpeasyVariable) -> pds.DataFrame:
    """Convert a :class:`~speasy.common.variable.SpeasyVariable` to pandas.DataFrame.

    See Also
    --------
    SpeasyVariable.to_dataframe
    """
    return SpeasyVariable.to_dataframe(var)


def merge(variables: List[SpeasyVariable]) -> Optional[SpeasyVariable]:
    """Merge a list of :class:`~speasy.common.variable.SpeasyVariable` objects.

    Parameters
    ----------
    variables: List[SpeasyVariable]
        Variables to merge together

    Returns
    -------
    SpeasyVariable:
        Resulting variable from merge operation
    """
    if len(variables) == 0:
        return None
    sorted_var_list = [v for v in variables if (
        v is not None) and (len(v.time) > 0)]
    sorted_var_list.sort(key=lambda v: v.time[0])

    # drop variables covered by previous ones
    for prev, current in zip(sorted_var_list[:-1], sorted_var_list[1:]):
        if prev.time[-1] >= current.time[-1]:
            sorted_var_list.remove(current)

    # drop variables covered by next ones
    for current, nxt in zip(sorted_var_list[:-1], sorted_var_list[1:]):
        if nxt.time[0] == current.time[0] and nxt.time[-1] >= current.time[-1]:
            sorted_var_list.remove(current)

    if len(sorted_var_list) == 0:
        for v in variables:
            if v is not None:
                return SpeasyVariable.reserve_like(v, length=0)
        return None

    overlaps = [
        np.where(current.time >= nxt.time[0])[0][0]
        if current.time[-1] >= nxt.time[0]
        else -1
        for current, nxt in zip(sorted_var_list[:-1], sorted_var_list[1:])
    ]

    dest_len = int(
        np.sum(
            [
                overlap if overlap != -1 else len(r.time)
                for overlap, r in zip(overlaps, sorted_var_list[:-1])
            ]
        )
    )
    dest_len += len(sorted_var_list[-1].time)

    result = SpeasyVariable.reserve_like(sorted_var_list[0], dest_len)

    pos = 0

    for r, overlap in zip(sorted_var_list, overlaps + [-1]):
        frag_len = len(r.time) if overlap == -1 else overlap
        result[pos: (pos + frag_len)] = r[0:frag_len]
        pos += frag_len
    return result



import logging
import time
from enum import Enum

from speasy.core import http, pack_kwargs
from speasy.core.cache import CacheCall
from speasy.config import amda as amda_cfg

import xml.etree.ElementTree as Et

from typing import Dict

log = logging.getLogger(__name__)

AMDA_BATCH_MODE_TIME = 240  # seconds


class Endpoint(Enum):
    """AMDA_Webservice REST API endpoints.
    """
    AUTH = "auth.php"
    OBSTREE = "getObsDataTree.php"

    LISTTT = "getTimeTablesList.php"
    LISTCAT = "getCatalogsList.php"
    LISTPARAM = "getParameterList.php"

    GETTT = "getTimeTable.php"
    GETCAT = "getCatalog.php"
    GETPARAM = "getParameter.php"

    GETSTATUS = "getStatus.php"


def auth_args(username: str, password: str) -> dict:
    return {'userID': username, 'password': password}


def request_url(endpoint: Endpoint, server_url: str) -> str:
    """Generates full URL for the given endpoint.

    Parameters
    ----------
    endpoint: Endpoint
        target API endpoint
    server_url: str
        server base url

    Returns
    -------
    str
        full URL to perform a request on the given API endpoint
    """
    if isinstance(endpoint, Endpoint):
        return f"{server_url}/php/rest/{endpoint.value}"
    else:
        raise TypeError(f"You must provide an {Endpoint} instead of {type(endpoint)}")


def token(server_url: str = amda_cfg.entry_point()) -> str:
    """Returns authentication token.

    Parameters
    ----------
    server_url:str
        server base URL on which the API token will be generated

    Returns
    -------
    str
        the generated token
    """
    # url = "{0}/php/rest/auth.php?".format(self.server_url)
    r = http.get(request_url(Endpoint.AUTH, server_url=server_url))
    if r.status_code == 200:
        return r.text.strip()
    else:
        raise RuntimeError("Failed to get auth token")


def send_request(endpoint: Endpoint, params: dict = None, timeout: int = http.DEFAULT_TIMEOUT,
                 server_url: str = amda_cfg.entry_point()) -> str or None:
    """Send a request on the AMDA_Webservice REST service to the given endpoint with given parameters.

    Parameters
    ----------
    endpoint: Endpoint
        target API endpoint on which the request will be performed
    params: dict
        request parameters
    timeout: int
        request timeout
    server_url: str
        the base server URL

    Returns
    -------
    str or None
        request result text, stripped of spaces and newlines

    """
    url = request_url(endpoint, server_url=server_url)
    params = params or {}
    params['token'] = token(server_url=server_url)
    r = http.get(url, params=params, timeout=timeout)
    if r.status_code == 200:
        return r.text.strip()
    return None


def send_indirect_request(endpoint: Endpoint, params: dict = None,
                          timeout: int = http.DEFAULT_TIMEOUT,
                          server_url: str = amda_cfg.entry_point()) -> str or None:
    """Send a request on the AMDA_Webservice REST service to the given endpoint with given parameters.
    The request is special in that the result is the URL to an XML file containing
    the actual data we are interested in. That is why we call :data:`requests.get()` twice in a row.

    Parameters
    ----------
    endpoint: Endpoint
        target API endpoint on which the request will be performed
    params: dict
        request parameters
    timeout: int
        request timeout
    server_url: str
        the base server URL

    Returns
    -------
    str or None
        request result text, stripped of spaces and newlines

    """
    next_url = send_request(endpoint=endpoint, params=params, timeout=timeout, server_url=server_url)
    if '<' in next_url and '>' in next_url:
        next_url = next_url.split(">")[1].split("<")[0]
    r = http.get(next_url, timeout=timeout)
    if r.status_code == 200:
        return r.text.strip()
    return None


def send_request_json(endpoint: Endpoint, params: Dict = None, timeout: int = http.DEFAULT_TIMEOUT,
                      server_url: str = amda_cfg.entry_point(),
                      extra_http_headers: Dict or None = None) -> str or None:
    """Send a request on the AMDA_Webservice REST service to the given endpoint with given parameters.
    We expect the result to be JSON data.

    Parameters
    ----------
    endpoint: Endpoint
        target API endpoint on which the request will be performed
    params: dict
        request parameters
    timeout: int
        request timeout
    server_url: str
        the base server URL

    Returns
    -------
    str or None
        request result parsed as json object
    """

    url = request_url(endpoint, server_url=server_url)
    params = params or {}
    http_headers = extra_http_headers or {}
    params['token'] = token(server_url=server_url)
    r = http.get(url, params=params, headers=http_headers, timeout=timeout)
    js = r.json()
    if 'success' in js and \
       js['success'] is True and \
       'dataFileURLs' in js:
        log.debug(f"success: {js['dataFileURLs']}")
        return js['dataFileURLs']
    elif "success" in js and \
         js["success"] is True and \
         "status" in js and \
         js["status"] == "in progress":
        log.warning("This request duration is too long, consider reducing time range")
        while True:
            default_sleep_time = 10.
            time.sleep(default_sleep_time)
            url = request_url(Endpoint.GETSTATUS, server_url=server_url)

            status = http.get(url, params=js, headers=http_headers).json()
            if status is not None and status["status"] == "done":
                return status["dataFileURLs"]
    else:
        log.debug(f"Failed: {r.text}")
    return None


@CacheCall(cache_retention=amda_cfg.user_cache_retention(), is_pure=True)
def get_timetables_xml_tree(server_url: str = amda_cfg.entry_point(), **kwargs: Dict) -> str or None:
    """Get list of timetables.

    Parameters
    ----------
    server_url: str
        the base server URL
    kwargs: dict
        extra request arguments such as username and password for private timetables

    Returns
    -------
    str or None
        request result, XML formatted text

    """
    return send_indirect_request(Endpoint.LISTTT, params=kwargs, server_url=server_url)


@CacheCall(cache_retention=amda_cfg.user_cache_retention(), is_pure=True)
def get_catalogs_xml_tree(server_url: str = amda_cfg.entry_point(), **kwargs: Dict) -> str or None:
    """Get list of catalogs.

    Parameters
    ----------
    server_url: str
        the base server URL
    kwargs: dict
        extra request arguments such as username and password for private catalogs

    Returns
    -------
    str or None
        request result, XML formatted text
    """
    return send_indirect_request(Endpoint.LISTCAT, params=kwargs, server_url=server_url)


@CacheCall(cache_retention=amda_cfg.user_cache_retention(), is_pure=True)
def get_user_timetables_xml_tree(username: str, password: str, server_url: str = amda_cfg.entry_point(),
                                 **kwargs: Dict) -> str or None:
    """Get private list of timetables.

    Parameters
    ----------
    username: str
        AMDA username
    password:
        AMDA password
    server_url: str
        the base server URL
    kwargs: dict
        extra request arguments

    Returns
    -------
    str or None
        request result, XML formatted text
    """
    return get_timetables_xml_tree(**auth_args(username, password), **kwargs, server_url=server_url)


@CacheCall(cache_retention=amda_cfg.user_cache_retention(), is_pure=True)
def get_user_catalogs_xml_tree(username: str, password: str, server_url: str = amda_cfg.entry_point(),
                               **kwargs: Dict) -> str or None:
    """Get private list of catalogs.

    Parameters
    ----------
    username: str
        AMDA username
    password:
        AMDA password
    server_url: str
        the base server URL
    kwargs: dict
        extra request arguments

    Returns
    -------
    str or None
        request result, XML formatted text
    """
    return get_catalogs_xml_tree(**auth_args(username, password), **kwargs, server_url=server_url)


@CacheCall(cache_retention=amda_cfg.user_cache_retention(), is_pure=True)
def get_user_parameters_xml_tree(username: str, password: str, server_url: str = amda_cfg.entry_point(),
                                 **kwargs: Dict) -> str or None:
    """Get private list of parameters.

    Parameters
    ----------
    username: str
        AMDA username
    password:
        AMDA password
    server_url: str
        the base server URL
    kwargs: dict
        extra request arguments

    Returns
    -------
    str or None
        request result, XML formatted text
    """
    xml_resp = send_request(Endpoint.LISTPARAM, params=pack_kwargs(**kwargs, **auth_args(username, password)),
                            server_url=server_url).strip()
    node = Et.fromstring(f'<root>{xml_resp}</root>').find("UserDefinedParameters")
    if node is not None:
        return http.get(node.text).text
    return None


@CacheCall(cache_retention=amda_cfg.user_cache_retention(), is_pure=True)
def get_timetable(server_url: str = amda_cfg.entry_point(), **kwargs: Dict) -> str or None:
    """Get timetable request.

    Parameters
    ----------
    server_url: str
        the base server URL
    kwargs: dict
        extra request arguments such as username and password for private timetables

    Returns
    -------
    str or None
        request result, XML formatted text
    """
    return send_request(Endpoint.GETTT, params=kwargs, server_url=server_url)


@CacheCall(cache_retention=amda_cfg.user_cache_retention(), is_pure=True)
def get_catalog(server_url: str = amda_cfg.entry_point(), **kwargs: Dict) -> str or None:
    """Get catalog request.

    Parameters
    ----------
    server_url: str
        the base server URL
    kwargs: dict
        extra request arguments such as username and password for private catalogs

    Returns
    -------
    str or None
        request result, XML formatted text
    """

    return send_request(Endpoint.GETCAT, params=kwargs, server_url=server_url)


def get_parameter(server_url: str = amda_cfg.entry_point(), extra_http_headers: Dict or None = None,
                  **kwargs: Dict) -> str or None:
    """Get parameter request.

    Parameters
    ----------
    extra_http_headers : Dict or None
        reserved for internal use
    server_url: str
        the base server URL
    kwargs: dict
        extra request arguments such as username and password for private parameters

    Returns
    -------
    str or None
        request result, XML formatted text
    """
    return send_request_json(Endpoint.GETPARAM, params=kwargs, server_url=server_url, timeout=AMDA_BATCH_MODE_TIME+10,
                             extra_http_headers=extra_http_headers)


@CacheCall(cache_retention=24 * 60 * 60, is_pure=True)
def get_obs_data_tree(server_url: str = amda_cfg.entry_point()) -> str or None:
    """Get observatory data tree.

    Parameters
    ----------
    server_url: str
        the base server URL

    Returns
    -------
    str or None
        request result, XML formatted text
    """
    return send_indirect_request(Endpoint.OBSTREE, server_url=server_url)



from speasy import SpeasyVariable
from .cache import CacheItem
from typing import List, Tuple
from speasy.core.datetime_range import DateTimeRange
from speasy.core import progress_bar
from speasy.products.variable import merge as merge_variables, to_dictionary, from_dictionary
from speasy.core.inventory.indexes import ParameterIndex
from datetime import datetime, timedelta
from functools import wraps
import logging
import math
from ._instance import _cache

log = logging.getLogger(__name__)

CACHE_ALLOWED_KWARGS = ['disable_cache']


def lower_hour_bound(dt: datetime, factor: int):
    return math.floor(dt.hour / factor) * factor


def upper_hour_bound(dt: datetime, factor: int):
    offsef = int(bool(dt - dt.replace(minute=0, second=0, microsecond=0)))
    return max(math.ceil((dt.hour + offsef) / factor), 1) * factor


def round_for_cache(dt_range: DateTimeRange, fragment_hours: int):
    start_time = dt_range.start_time.replace(hour=lower_hour_bound(dt_range.start_time, fragment_hours), minute=0,
                                             second=0, microsecond=0)
    stop_time = dt_range.stop_time.replace(hour=0, minute=0, second=0, microsecond=0) + timedelta(
        hours=upper_hour_bound(dt_range.stop_time, fragment_hours))
    return DateTimeRange(start_time, stop_time)


def is_up_to_date(item: CacheItem, version):
    return (item.version is None) or (item.version >= version)


def group_fragments_if(fragments, predicate):
    merged = []
    if len(fragments):
        merged = [[fragments[0]]]
        for fragment in fragments[1:]:
            if predicate(merged[-1][-1], fragment):
                merged[-1].append(fragment)
            else:
                merged.append([fragment])
    return merged


def group_contiguous_fragments(fragments, duration):
    return group_fragments_if(fragments, lambda previous, current: (previous + duration * 1.01) > current)


def default_cache_entry_name(prefix: str, product: str, start_time: str, **kwargs):
    return f"{prefix}/{product}/{start_time}"


def product_name(product: str or ParameterIndex):
    if type(product) is str:
        return product
    elif isinstance(product, ParameterIndex):
        return product.spz_uid()
    else:
        raise TypeError(f'Product must either be str or ParameterIndex got {type(product)}')


class _Cacheable:
    def __init__(self, prefix, cache_instance=_cache, start_time_arg='start_time', stop_time_arg='stop_time',
                 version=None,
                 fragment_hours=lambda x: 1, cache_margins=1.2, leak_cache=False, entry_name=default_cache_entry_name):
        self.start_time_arg = start_time_arg
        self.stop_time_arg = stop_time_arg
        self.version = (lambda x, y: 0) if version is None else version
        self.fragment_hours = fragment_hours
        self.cache_margins = cache_margins
        self.cache = cache_instance
        self.prefix = prefix
        self.leak_cache = leak_cache
        self.entry_name = entry_name

    def add_to_cache(self, variable: SpeasyVariable or None, fragments, product, fragment_duration_hours, version,
                     **kwargs) -> SpeasyVariable or None:
        if variable is not None:
            for fragment in fragments:
                self.set_cache_entry(fragment, product,
                                     CacheItem(to_dictionary(
                                         variable[fragment:(fragment + timedelta(hours=fragment_duration_hours))]),
                                         version))
        return variable

    def set_cache_entry(self, fragment, product: str, entry, **kwargs):
        key = self.entry_name(self.prefix, product, fragment.isoformat(), **kwargs)
        log.debug(f"add {key} into cache")
        self.cache[key] = entry

    def get_cache_entry(self, fragment: datetime, product, **kwargs):
        key = self.entry_name(self.prefix, product, fragment.isoformat(), **kwargs)
        if key in self.cache:
            entry = self.cache[key]
            log.debug(f"Found {key} inside cache")
            return entry
        else:
            log.debug(f"{key} not found inside cache")
        return None

    def get_from_cache(self, fragment, product, version, **kwargs):
        entry = self.get_cache_entry(fragment, product, **kwargs)
        if entry is not None:
            if is_up_to_date(entry, version):
                try:
                    return from_dictionary(entry.data)
                except Exception as e:
                    log.warning(f"got an exception {e} while loading fragment {fragment} for {product}")
                    return None
            log.debug(f"Cache entry is outdated")
        return None

    def fragment_list(self, product, dt_range) -> Tuple[int, List[datetime]]:
        fragment_hours = self.fragment_hours(product)
        cache_dt_range = round_for_cache(dt_range * self.cache_margins, fragment_hours)
        dt = timedelta(hours=fragment_hours)
        fragments = [cache_dt_range.start_time + i * dt for i in range(math.ceil(cache_dt_range.duration / dt))]
        return fragment_hours, fragments

    def get_fragments_from_cache(self, fragments: List[datetime], product: str, version, **kwargs):
        data_fragments = []
        with self.cache.transact():
            for fragment in fragments:
                data_fragments.append(self.get_from_cache(fragment, product, version, **kwargs))
        return data_fragments

    def get_cache_entries(self, fragments: List[datetime], product: str, **kwargs):
        return [self.get_cache_entry(fragment, product, **kwargs) for fragment in fragments]


class Cacheable(object):
    def __init__(self, prefix, cache_instance=_cache, start_time_arg='start_time', stop_time_arg='stop_time',
                 version=None,
                 fragment_hours=lambda x: 1, cache_margins=1.2, leak_cache=False, entry_name=default_cache_entry_name):
        self._cache = _Cacheable(prefix, cache_instance=cache_instance, start_time_arg=start_time_arg,
                                 stop_time_arg=stop_time_arg,
                                 version=version,
                                 fragment_hours=fragment_hours, cache_margins=cache_margins, leak_cache=leak_cache,
                                 entry_name=entry_name)

    def __call__(self, get_data):
        @wraps(get_data)
        def wrapped(wrapped_self, product, start_time, stop_time, **kwargs):
            product = product_name(product)
            version = self._cache.version(wrapped_self, product)
            dt_range = DateTimeRange(start_time, stop_time)
            if kwargs.pop("disable_cache", False):
                return get_data(wrapped_self, product=product, start_time=dt_range.start_time,
                                stop_time=dt_range.stop_time, **kwargs)

            fragment_hours, fragments = self._cache.fragment_list(product, dt_range)
            fragment_duration = timedelta(hours=fragment_hours)
            data_chunks = self._cache.get_fragments_from_cache(fragments=fragments, product=product, version=version,
                                                               **kwargs)
            missing_fragments = group_contiguous_fragments(
                [fragment for f_data, fragment in zip(data_chunks, fragments) if f_data is None],
                duration=fragment_duration)

            data_chunks += [
                self._cache.add_to_cache(
                    get_data(
                        wrapped_self, product=product, start_time=fragment_group[0],
                        stop_time=fragment_group[-1] + fragment_duration, **kwargs),
                    fragments=fragment_group, product=product, fragment_duration_hours=fragment_hours,
                    version=version, **kwargs)
                for fragment_group
                in
                progress_bar(leave=False, desc="Downloading missing fragments from cache", **kwargs)(missing_fragments)]

            data_chunks = list(filter(lambda d: d is not None, data_chunks))

            if len(data_chunks):
                if len(data_chunks) == 1:
                    return data_chunks[0][dt_range.start_time:dt_range.stop_time].copy()
                data_chunks[0] = data_chunks[0][dt_range.start_time:]
                data_chunks[-1] = data_chunks[-1][:dt_range.stop_time]
                return merge_variables(data_chunks)[dt_range.start_time:dt_range.stop_time]
            return None

        if self._cache.leak_cache:
            wrapped.cache = self._cache.cache
        return wrapped


class UnversionedProviderCache(object):
    def __init__(self, prefix, cache_instance=_cache, start_time_arg='start_time', stop_time_arg='stop_time',
                 fragment_hours=lambda x: 1, cache_margins=1.2, leak_cache=False, entry_name=default_cache_entry_name,
                 cache_retention=None):
        self._cache = _Cacheable(prefix, cache_instance=cache_instance, start_time_arg=start_time_arg,
                                 stop_time_arg=stop_time_arg,
                                 version=lambda x, y: datetime.utcnow().isoformat(),
                                 fragment_hours=fragment_hours, cache_margins=cache_margins, leak_cache=leak_cache,
                                 entry_name=entry_name)
        self.cache_retention = cache_retention or timedelta(days=14)

    def split_fragments(self, fragments, product, fragment_duration, **kwargs):
        entries = self._cache.get_cache_entries(fragments=fragments, product=product, **kwargs)
        missing_fragments = []
        data_chunks = []
        maybe_outdated_fragments = []
        for fragment, entry in zip(fragments, entries):
            if entry is None:
                missing_fragments.append(fragment)
            elif (entry.version + self.cache_retention) > datetime.utcnow():
                try:
                    data_chunks.append(from_dictionary(entry.data))
                except Exception as e:
                    missing_fragments.append(fragment)
                    log.warning(f"got an exception {e} while loading fragment {fragment} for {product}")
            else:
                maybe_outdated_fragments.append((fragment, entry))

        missing_fragments = group_contiguous_fragments(missing_fragments, duration=fragment_duration)
        # This is a deliberate choice here to group fragments in order to reduce requests count, the bet here is
        # that it costs less to asks for more data in one shot then doing several requests. To be more clear about
        # the issue here, grouping fragments implies choosing a date to compare for the whole group and by
        # construction each fragment inside the group is likely to have different date. So the safe choice is to
        # declare the whole group as old as the oldest element which leads to maybe updating some fragments inside
        # the group that were up-to-date.
        maybe_outdated_fragments = group_fragments_if(
            maybe_outdated_fragments,
            lambda previous, current: (previous[0] + fragment_duration * 1.01) > current[0])
        return data_chunks, maybe_outdated_fragments, missing_fragments

    def __call__(self, get_data):
        @wraps(get_data)
        def wrapped(wrapped_self, product, start_time, stop_time, **kwargs):
            product = product_name(product)
            dt_range = DateTimeRange(start_time, stop_time)
            if kwargs.pop("disable_cache", False):
                return get_data(wrapped_self, product=product, start_time=dt_range.start_time,
                                stop_time=dt_range.stop_time, **kwargs)

            fragment_hours, fragments = self._cache.fragment_list(product, dt_range)
            fragment_duration = timedelta(hours=fragment_hours)
            data_chunks, maybe_outdated_fragments, missing_fragments = self.split_fragments(fragments, product,
                                                                                            fragment_duration, **kwargs)
            data_chunks += \
                list(filter(lambda d: d is not None, [
                    self._cache.add_to_cache(
                        get_data(
                            wrapped_self, product=product, start_time=fragment_group[0],
                            stop_time=fragment_group[-1] + fragment_duration, **kwargs),
                        fragments=fragment_group, product=product, fragment_duration_hours=fragment_hours,
                        version=datetime.utcnow(), **kwargs)
                    for fragment_group
                    in progress_bar(leave=False, desc="Downloading missing fragments from cache", **kwargs)(
                        missing_fragments)]))

            for group in progress_bar(leave=False, desc="Checking if cache fragments are outdated", **kwargs)(
                maybe_outdated_fragments):
                oldest = max(group, key=lambda item: item[1].version)[1].version
                data = get_data(wrapped_self, product=product, start_time=group[0][0],
                                stop_time=group[-1][0] + fragment_duration, if_newer_than=oldest, **kwargs)
                if data is None:
                    for fragment, entry in group:
                        entry.version = datetime.utcnow()
                        self._cache.set_cache_entry(fragment, product, entry)
                        data_chunks.append(entry.data)
                else:
                    self._cache.add_to_cache(data, [item[0] for item in group], product,
                                             fragment_duration_hours=fragment_hours,
                                             version=datetime.now(), **kwargs)
                    data_chunks.append(data)

            if len(data_chunks):
                if len(data_chunks) == 1:
                    return data_chunks[0][dt_range.start_time:dt_range.stop_time].copy()
                data_chunks[0] = data_chunks[0][dt_range.start_time:]
                data_chunks[-1] = data_chunks[-1][:dt_range.stop_time]
                return merge_variables(data_chunks)[dt_range.start_time:dt_range.stop_time]
            return None

        if self._cache.leak_cache:
            wrapped.cache = self._cache.cache
        return wrapped



from datetime import datetime, timedelta
from speasy.core import span_utils, make_utc_datetime
from typing import List
import numpy as np


class DateTimeRange:
    __slots__ = ['_rng']

    def __init__(self, start_time: datetime or str or np.float64 or float or np.datetime64,
                 stop_time: datetime or str or np.float64 or float or np.datetime64):
        self._rng = [make_utc_datetime(start_time), make_utc_datetime(stop_time)]

    @property
    def start_time(self) -> datetime:
        return self._rng[0]

    @start_time.setter
    def start_time(self, start_time: datetime or str or np.float64 or float or np.datetime64):
        self._rng[0] = make_utc_datetime(start_time)

    @property
    def stop_time(self) -> datetime:
        return self._rng[1]

    @stop_time.setter
    def stop_time(self, stop_time: datetime or str or np.float64 or float or np.datetime64):
        self._rng[1] = make_utc_datetime(stop_time)

    @property
    def duration(self) -> timedelta:
        return self.stop_time - self.start_time

    def split(self, fragment_duration: timedelta) -> List["DateTimeRange"]:
        return span_utils.split(self, fragment_duration)

    def __eq__(self, other) -> bool:
        return span_utils.equals(self, other)

    def intersect(self, other):
        return span_utils.intersects(self, other)

    def __repr__(self):
        return f'<DateTimeRange: {self.start_time.isoformat()} -> {self.stop_time.isoformat()}>'

    def __getitem__(self, item):
        return self._rng.__getitem__(item)

    def __setitem__(self, key, value):
        self._rng.__setitem__(key, value)

    def __len__(self):
        return 2

    def __contains__(self, item: object) -> bool:
        return span_utils.contains(self, other=item)

    def __add__(self, other):
        if type(other) is timedelta:
            return span_utils.shift(self, other)
        else:
            raise TypeError()

    def __sub__(self, other):
        if type(other) is timedelta:
            return span_utils.shift(self, -other)
        else:
            return span_utils.difference(self, other)

    def __mul__(self, other):
        return span_utils.zoom(self, factor=other)



from speasy import __version__
import platform
import requests
import socket
from requests.utils import quote as _quote
from requests.adapters import HTTPAdapter
from requests.packages.urllib3.util.retry import Retry
from urllib.request import Request, urlopen, HTTPError, URLError
from time import sleep
import logging

log = logging.getLogger(__name__)

USER_AGENT = f'Speasy/{__version__} {platform.uname()} (SciQLop project)'

DEFAULT_TIMEOUT = 60  # seconds

DEFAULT_DELAY = 5  # seconds

DEFAULT_RETRY_COUNT = 5

STATUS_FORCE_LIST = [500, 502, 504]

RETRY_AFTER_LIST = [429, 503]  # Note: Specific treatment for 429 & 503 error codes (see below)


class TimeoutHTTPAdapter(HTTPAdapter):
    def __init__(self, *args, timeout=DEFAULT_TIMEOUT, **kwargs):
        self.timeout = timeout
        super().__init__(*args, **kwargs)

    def send(self, request, **kwargs):
        kwargs.pop('timeout', None)
        return super().send(request, timeout=self.timeout, **kwargs)


def quote(*args, **kwargs):
    return _quote(*args, **kwargs)


def apply_delay(headers: dict = None):
    delay = DEFAULT_DELAY
    try:
        if headers and ('Retry-After' in headers):
            delay = float(headers['Retry-After'])
    except ValueError:
        pass
    log.debug(f"Will sleep for {delay} seconds")
    sleep(delay)


def get(url, headers: dict = None, params: dict = None, timeout: int = DEFAULT_TIMEOUT, head_only: bool = False):
    headers = {} if headers is None else headers
    headers['User-Agent'] = USER_AGENT
    # cf. https://findwork.dev/blog/advanced-usage-python-requests-timeouts-retries-hooks/
    retry_strategy = Retry(
        total=DEFAULT_RETRY_COUNT,
        backoff_factor=1,
        status_forcelist=STATUS_FORCE_LIST,
        allowed_methods=["HEAD", "GET"]
    )
    adapter = TimeoutHTTPAdapter(max_retries=retry_strategy, timeout=timeout)
    http = requests.Session()
    http.mount("https://", adapter)
    http.mount("http://", adapter)
    while True:
        if head_only:
            resp = http.head(url, headers=headers, params=params)
        else:
            resp = http.get(url, headers=headers, params=params)
        if resp.status_code in RETRY_AFTER_LIST:  # Honor "Retry-After"
            log.debug(f"Got {resp.status_code} response")
            apply_delay(resp.headers)
        else:
            break
    return resp


def urlopen_with_retry(url, timeout: int = DEFAULT_TIMEOUT, headers: dict = None):
    headers = {} if headers is None else headers
    headers['User-Agent'] = USER_AGENT
    req = Request(url, headers=headers)
    retrycount = 0
    while True:
        try:
            resp = urlopen(req, timeout=timeout)
            return resp
        except HTTPError as e:
            if isinstance(e.reason, socket.timeout):
                log.debug("Timeout exception during urlopen request")
            elif e.code in STATUS_FORCE_LIST:
                log.debug(f"HTTP Error Got {e.code} response")
            elif e.code in RETRY_AFTER_LIST:  # Honor "Retry-After"
                log.debug(f"Got {e.code} response")
                apply_delay(resp.headers)
                return urlopen_with_retry(url, timeout=timeout, headers=headers)
            else:
                raise e
            retrycount += 1
            if retrycount > DEFAULT_RETRY_COUNT:
                raise e
            apply_delay()
        except URLError as e:
            log.debug(f"Got {e.reason} error")
            retrycount += 1
            if retrycount > DEFAULT_RETRY_COUNT:
                raise e
            apply_delay()



from .cache import Cache, CacheItem
from ._function_cache import CacheCall
from ._providers_caches import CACHE_ALLOWED_KWARGS, Cacheable, UnversionedProviderCache
from ._instance import _cache


def cache_len():
    return len(_cache)


def cache_disk_size():
    return _cache.disk_size()


def stats():
    return _cache.stats()


def entries():
    return _cache.keys()


def add_item(key, item, expires=None):
    _cache.set(key, item, expires)


def get_item(key, default_value=None):
    return _cache.get(key, default_value)



import requests
from astroquery.utils.tap.core import TapPlus
from typing import Optional, Tuple, Dict
from datetime import datetime, timedelta
from speasy.core.cache import Cacheable, CACHE_ALLOWED_KWARGS  # _cache is used for tests (hack...)
from speasy.products.variable import SpeasyVariable
from speasy.core import http, AllowedKwargs, fix_name
from speasy.core.proxy import Proxyfiable, GetProduct, PROXY_ALLOWED_KWARGS
from speasy.core.cdf import load_variable
from speasy.core.inventory.indexes import ParameterIndex, DatasetIndex, SpeasyIndex, make_inventory_node
from speasy.core.dataprovider import DataProvider, ParameterRangeCheck, GET_DATA_ALLOWED_KWARGS
from tempfile import TemporaryDirectory
from speasy.core.datetime_range import DateTimeRange
from speasy.core.requests_scheduling import SplitLargeRequests
import tarfile
import logging
from io import BytesIO

log = logging.getLogger(__name__)


def to_dataset_and_variable(index_or_str: ParameterIndex or str) -> Tuple[str, str]:
    if type(index_or_str) is str:
        parts = index_or_str.split('/')
    elif type(index_or_str) is ParameterIndex:
        parts = index_or_str.product.split('/')
    else:
        raise TypeError(f"given parameter {index_or_str} of type {type(index_or_str)} is not a compatible index")
    assert len(parts) == 2
    return parts[0], parts[1]


def register_dataset(instruments, datasets, dataset):
    meta = {cname: dataset[cname] for cname in dataset.colnames}
    meta['stop_date'] = meta.pop('end_date')
    name = fix_name(meta['dataset_id'])
    node = make_inventory_node(instruments[dataset['instruments']], DatasetIndex, name=name,
                               provider="csa",
                               uid=meta['dataset_id'], **meta)
    datasets[meta['dataset_id']] = node


def register_observatory(missions, observatories, observatory):
    meta = {cname: observatory[cname] for cname in observatory.colnames}
    name = meta.pop('name')
    node = make_inventory_node(missions[observatory['mission_name']], SpeasyIndex,
                               name=fix_name(name),
                               provider="csa",
                               uid=name,
                               **meta)
    observatories[name] = node


def register_mission(inventory_tree, missions, mission):
    meta = {cname: mission[cname] for cname in mission.colnames}
    name = meta.pop('name')
    node = make_inventory_node(inventory_tree, SpeasyIndex, name=fix_name(name),
                               provider="csa",
                               uid=name, **meta)
    missions[name] = node


def register_instrument(observatories, instruments, instrument):
    meta = {cname: instrument[cname] for cname in instrument.colnames}
    name = meta.pop('name')
    node = make_inventory_node(observatories.get(instrument['observatories'], observatories['MULTIPLE']),
                               SpeasyIndex, name=fix_name(name),
                               provider="csa",
                               uid=name, **meta)
    instruments[name] = node


def register_param(datasets, parameter):
    parent_dataset = datasets.get(parameter["dataset_id"], None)
    if parent_dataset is not None:
        meta = {cname: parameter[cname] for cname in parameter.colnames}
        meta['dataset'] = parameter["dataset_id"]
        meta['start_date'] = parent_dataset.start_date
        meta['stop_date'] = parent_dataset.stop_date
        name = fix_name(meta['parameter_id'])
        make_inventory_node(parent_dataset, ParameterIndex, name=name,
                            provider="csa", uid=f"{parameter['dataset_id']}/{parameter['parameter_id']}", **meta)


def build_inventory(root: SpeasyIndex, tapurl="https://csa.esac.esa.int/csa-sl-tap/tap/"):
    CSA = TapPlus(url=tapurl)
    missions_req = CSA.launch_job_async("SELECT * FROM csa.v_mission")
    observatories_req = CSA.launch_job_async("SELECT * FROM csa.v_observatory")
    instruments_req = CSA.launch_job_async("SELECT * FROM csa.v_instrument")
    datasets_req = CSA.launch_job_async("SELECT * FROM csa.v_dataset WHERE is_cef='true' AND is_istp='true'")
    parameters_req = CSA.launch_job_async("SELECT * FROM csa.v_parameter WHERE data_type='Data'")
    missions = {}
    observatories = {}
    instruments = {}
    datasets = {}
    list(map(lambda m: register_mission(root, missions, m), missions_req.get_results()))
    list(map(lambda o: register_observatory(missions, observatories, o), observatories_req.get_results()))
    list(map(lambda i: register_instrument(observatories, instruments, i), instruments_req.get_results()))
    list(map(lambda d: register_dataset(instruments, datasets, d), datasets_req.get_results()))
    list(map(lambda p: register_param(datasets, p), parameters_req.get_results()))

    return root


def _read_cdf(response: requests.Response, variable: str) -> SpeasyVariable:
    with tarfile.open(fileobj=BytesIO(response.content)) as tar:
        tarname = tar.getnames()
        if len(tarname):
            with TemporaryDirectory() as tmp_dir:
                tar.extractall(tmp_dir)
                return load_variable(file=f"{tmp_dir}/{tarname[0]}", variable=variable)


def get_parameter_args(start_time: datetime, stop_time: datetime, product: str, **kwargs):
    return {'path': f"csa/{product}", 'start_time': f'{start_time.isoformat()}',
            'stop_time': f'{stop_time.isoformat()}'}


class CSA_Webservice(DataProvider):
    def __init__(self):
        DataProvider.__init__(self, provider_name='csa')
        self.__url = "https://csa.esac.esa.int/csa-sl-tap/data"

    def _dataset_range(self, dataset: str or DatasetIndex) -> DateTimeRange:
        if type(dataset) is str:
            dataset = self.flat_inventory.datasets[dataset]
        return DateTimeRange(dataset.start_date, dataset.stop_date)

    def _dl_variable(self,
                     dataset: str, variable: str,
                     start_time: datetime, stop_time: datetime, extra_http_headers: Dict[str, str] or None = None) -> \
        Optional[SpeasyVariable]:

        # https://csa.esac.esa.int/csa-sl-tap/data?RETRIEVAL_TYPE=product&&DATASET_ID=C3_CP_PEA_LERL_DEFlux&START_DATE=2001-06-10T22:12:14Z&END_DATE=2001-06-11T06:12:14Z&DELIVERY_FORMAT=CDF_ISTP&DELIVERY_INTERVAL=all
        ds_range = self._dataset_range(dataset)
        if not ds_range.intersect(DateTimeRange(start_time, stop_time)):
            log.warning(f"You are requesting {dataset}/{variable} outside of its definition range {ds_range}")
            return None
        headers = {}
        if extra_http_headers is not None:
            headers.update(extra_http_headers)
        resp = http.get(self.__url, params={
            "RETRIEVAL_TYPE": "product",
            "DATASET_ID": dataset,
            "START_DATE": start_time.strftime('%Y-%m-%dT%H:%M:%SZ'),
            "END_DATE": stop_time.strftime('%Y-%m-%dT%H:%M:%SZ'),
            "DELIVERY_FORMAT": "CDF_ISTP",
            "DELIVERY_INTERVAL": "all"
        }, headers=headers)
        log.debug(f"{resp.url}")
        if resp.status_code != 200:
            raise RuntimeError(f'Failed to get data with request: {resp.url}, got {resp.status_code} HTTP response')
        if not resp.ok:
            return None
        return _read_cdf(resp, variable)

    @staticmethod
    def build_inventory(root: SpeasyIndex):
        return build_inventory(root)

    def parameter_range(self, parameter_id: str or ParameterIndex) -> Optional[DateTimeRange]:
        """Get product time range.

        Parameters
        ----------
        parameter_id: str or ParameterIndex
            parameter id

        Returns
        -------
        Optional[DateTimeRange]
            Data time range

        Examples
        --------

        >>> import speasy as spz
        >>> spz.csa.parameter_range("C3_CP_WBD_WAVEFORM_BM2/B__C3_CP_WBD_WAVEFORM_BM2")
        <DateTimeRange: 2001-03-07T17:45:22+00:00 -> ...>

        """
        return self._parameter_range(parameter_id)

    def dataset_range(self, dataset_id: str or DatasetIndex) -> Optional[DateTimeRange]:
        """Get product time range.

        Parameters
        ----------
        dataset_id: str or DatasetIndex
            parameter id

        Returns
        -------
        Optional[DateTimeRange]
            Data time range

        Examples
        --------

        >>> import speasy as spz
        >>> spz.csa.dataset_range("D2_CP_FGM_SPIN")
        <DateTimeRange: 2004-07-27T00:00:00+00:00 -> ...>

        """
        return self._dataset_range(dataset_id)

    def product_last_update(self, product: str or ParameterIndex):
        """Get date of last modification of dataset or parameter.

        Parameters
        ----------
        product: str or ParameterIndex
            product

        Returns
        -------
        str
            product last update date
        """
        dataset, variable = to_dataset_and_variable(product)
        return self.flat_inventory.datasets[dataset].date_last_update

    @AllowedKwargs(PROXY_ALLOWED_KWARGS + CACHE_ALLOWED_KWARGS + GET_DATA_ALLOWED_KWARGS)
    @ParameterRangeCheck()
    @Cacheable(prefix="csa", fragment_hours=lambda x: 12, version=product_last_update)
    @SplitLargeRequests(threshold=lambda: timedelta(days=7))
    @Proxyfiable(GetProduct, get_parameter_args)
    def get_data(self, product, start_time: datetime, stop_time: datetime,
                 extra_http_headers: Dict[str, str] or None = None):
        dataset, variable = to_dataset_and_variable(product)
        return self._dl_variable(start_time=start_time, stop_time=stop_time, dataset=dataset,
                                 variable=variable, extra_http_headers=extra_http_headers)

    def get_variable(self, dataset: str, variable: str, start_time: datetime or str, stop_time: datetime or str,
                     **kwargs) -> \
        Optional[SpeasyVariable]:
        return self.get_data(f"{dataset}/{variable}", start_time, stop_time, **kwargs)



from speasy.core import fix_name
from speasy.core.inventory.indexes import DatasetIndex, SpeasyIndex, make_inventory_node
import xml.etree.ElementTree as Et


def alias_rules(name):
    rules = {
        "AC": "ACE",
        "Parker Solar Probe (PSP)": "ParkerSolarProbe",
        "PSP": "ParkerSolarProbe",
        "mms1": "MMS1",
        "mms2": "MMS2",
        "mms3": "MMS3",
        "mms4": "MMS4",
    }
    return rules.get(name, name)


def description(node) -> str:
    desc_node = node.find('{cdas}description')
    if desc_node is not None:
        return desc_node.attrib['short']
    return ""


def extract_node(node, is_dataset=False):
    name = node.attrib["serviceprovider_ID"]
    n = {
        'name': fix_name(alias_rules(name)),
        'description': description(node)
    }
    n.update(node.attrib)
    if is_dataset:
        n['start_date'] = n.pop('timerange_start')
        n['stop_date'] = n.pop('timerange_stop')
        master_cd_node = node.find('{cdas}mastercdf')
        if master_cd_node is not None:
            n["mastercdf"] = master_cd_node.attrib["serviceprovider_ID"]
    return n


def register_dataset(inventory_tree, mission_group_node, observatory_node, instrument_node, dataset_node):
    observatory = extract_node(observatory_node)
    mission_group = extract_node(mission_group_node)
    if mission_group['name'] != observatory['name']:
        inventory_tree = make_inventory_node(inventory_tree, SpeasyIndex, provider="cda",
                                             uid=mission_group.get('serviceprovider_ID'), **mission_group)
    inventory_tree = make_inventory_node(inventory_tree, SpeasyIndex, provider="cda",
                                         uid=observatory.get('serviceprovider_ID'), **observatory)
    if instrument_node.attrib["serviceprovider_ID"] != "":
        inventory_tree = make_inventory_node(inventory_tree, SpeasyIndex, provider="cda",
                                             uid=instrument_node.get('serviceprovider_ID'), **extract_node(instrument_node))
    inventory_tree = make_inventory_node(inventory_tree, DatasetIndex, provider="cda",
                                         uid=dataset_node.get('serviceprovider_ID'),
                                         **extract_node(dataset_node, is_dataset=True))
    return inventory_tree


def has_master_cdf(node):
    master_cdf = node.find('{cdas}mastercdf')
    if master_cdf is not None:
        return True
    return None


def parse_dataset(inventory_tree, dataset_node):
    mission_group_node = dataset_node.find('{cdas}mission_group')
    observatory_node = dataset_node.find('{cdas}observatory')
    instrument_node = dataset_node.find('{cdas}instrument')
    if has_master_cdf(dataset_node):
        return register_dataset(inventory_tree, mission_group_node, observatory_node, instrument_node, dataset_node)
    else:
        print(f'Missing master CDF for {dataset_node.attrib["serviceprovider_ID"]}')


def load_xml_catalog(xml_file_path: str, root: SpeasyIndex or None = None):
    with open(xml_file_path) as xml_file:
        tree = Et.fromstring(xml_file.read())
        inventory_tree = root or SpeasyIndex(name='root', provider='cda', uid='cda_root')
        for site in tree.iter('{cdas}datasite'):
            if site.attrib['ID'] == 'CDAWeb_HTTPS':
                for node in site.iter('{cdas}dataset'):
                    parse_dataset(inventory_tree, node)
        return inventory_tree



"""AMDA_Webservice utility functions. This module defines some conversion functions specific to AMDA_Webservice, mainly
conversion procedures for parsing CSV and VOTable data.

"""
import logging
import datetime
import os
from typing import Dict, List
import tempfile

import numpy as np
import pandas as pds

from speasy.core import epoch_to_datetime64
from speasy.core.http import urlopen_with_retry
from speasy.core.datetime_range import DateTimeRange
from speasy.products.catalog import Catalog, Event
from speasy.products.timetable import TimeTable
from speasy.products.variable import (DataContainer, SpeasyVariable,
                                      VariableAxis, VariableTimeAxis)

log = logging.getLogger(__name__)

DATA_CHUNK_SIZE = 10485760


def _copy_data(csv, fd):
    content_length = 0
    chunk_size = -1
    if hasattr(csv, 'getheader'):
        content_length = csv.getheader('content-length')
        if content_length:
            content_length = int(content_length)
            chunk_size = DATA_CHUNK_SIZE
    size = 0
    while True:
        chunk = csv.read(chunk_size)
        if not chunk:
            break
        fd.write(chunk)
        size += len(chunk)
        if content_length:
            percent = int((size / content_length)*100)
            log.debug(f"Download data: {percent}%")
    fd.seek(0)
    return fd


def load_csv(filename: str) -> SpeasyVariable:
    """Load a CSV file

    Parameters
    ----------
    filename: str
        CSV filename

    Returns
    -------
    SpeasyVariable
        CSV contents
    """
    if '://' not in filename:
        filename = f"file:///{os.path.abspath(filename)}"
    with urlopen_with_retry(filename) as csv:
        with tempfile.TemporaryFile() as fd:
            _copy_data(csv, fd)
            line = fd.readline().decode()
            meta = {}
            y = None
            y_label = None
            while len(line) > 0 and line[0] == '#':
                if ':' in line:
                    key, value = line[1:].split(':', 1)
                    meta[key.strip()] = value.strip()
                line = fd.readline().decode()
            columns = [col.strip()
                       for col in meta.get('DATA_COLUMNS', "").split(', ')[:]]
            meta["UNITS"] = meta.get("PARAMETER_UNITS")
            fd.seek(0)
            data = pds.read_csv(fd, comment='#', delim_whitespace=True,
                                header=None, names=columns).values.transpose()
            time, data = epoch_to_datetime64(data[0]), data[1:].transpose()

        if "PARAMETER_TABLE_MIN_VALUES[1]" in meta:
            min_v = np.array(
                [float(v) for v in meta["PARAMETER_TABLE_MIN_VALUES[1]"].split(',')])
            max_v = np.array(
                [float(v) for v in meta["PARAMETER_TABLE_MAX_VALUES[1]"].split(',')])
            y_label = meta["PARAMETER_TABLE[1]"]
            y = (max_v + min_v) / 2.
        elif "PARAMETER_TABLE_MIN_VALUES[0]" in meta:
            min_v = np.array(
                [float(v) for v in meta["PARAMETER_TABLE_MIN_VALUES[0]"].split(',')])
            max_v = np.array(
                [float(v) for v in meta["PARAMETER_TABLE_MAX_VALUES[0]"].split(',')])
            y = (max_v + min_v) / 2.
            y_label = meta["PARAMETER_TABLE[0]"]
        time_axis = VariableTimeAxis(values=time)
        if y is None:
            axes = [time_axis]
        else:
            axes = [time_axis, VariableAxis(
                name=y_label, values=y, is_time_dependent=False)]
        return SpeasyVariable(
            axes=axes,
            values=DataContainer(values=data, meta=meta),
            columns=columns[1:])


def _build_event(data, colnames: List[str]) -> Event:
    return Event(datetime.datetime.strptime(data[0], "%Y-%m-%dT%H:%M:%S.%f"),
                 datetime.datetime.strptime(data[1], "%Y-%m-%dT%H:%M:%S.%f"),
                 {name: value for name, value in zip(colnames[2:], data[2:])})


def load_timetable(filename: str) -> TimeTable:
    """Load a timetable file

    Parameters
    ----------
    filename: str
        filename

    Returns
    -------
    TimeTable
        File content loaded as TimeTable
    """
    if '://' not in filename:
        filename = f"file://{os.path.abspath(filename)}"
    with urlopen_with_retry(filename) as votable:
        # save the timetable as a dataframe, speasy.common.SpeasyVariable
        # get header data first
        import io

        from astropy.io.votable import parse as parse_votable
        votable = parse_votable(io.BytesIO(votable.read()))
        name = next(filter(lambda e: 'Name' in e,
                           votable.description.split(';\n'))).split(':')[-1]
        # convert astropy votable structure to SpeasyVariable
        tab = votable.get_first_table()
        # prepare data
        data = tab.array.tolist()
        dt_ranges = [DateTimeRange(datetime.datetime.strptime(t0, "%Y-%m-%dT%H:%M:%S.%f"),
                                   datetime.datetime.strptime(t1, "%Y-%m-%dT%H:%M:%S.%f")) for (t0, t1) in
                     data]
        var = TimeTable(name=name, meta={}, dt_ranges=dt_ranges)
        return var


def load_catalog(filename: str) -> Catalog:
    """Load a timetable file

    Parameters
    ----------
    filename: str
        filename

    Returns
    -------
    Catalog
        File content loaded as Catalog

    """
    if '://' not in filename:
        filename = f"file://{os.path.abspath(filename)}"
    with urlopen_with_retry(filename) as votable:
        # save the timetable as a dataframe, speasy.common.SpeasyVariable
        # get header data first
        import io

        from astropy.io.votable import parse as parse_votable
        votable = parse_votable(io.BytesIO(votable.read()))
        # convert astropy votable structure to SpeasyVariable
        tab = votable.get_first_table()
        name = next(filter(lambda e: 'Name' in e,
                           votable.description.split(';\n'))).split(':')[-1]
        colnames = list(map(lambda f: f.name, tab.fields))
        data = tab.array.tolist()
        events = [_build_event(line, colnames) for line in data]
        var = Catalog(name=name, meta={}, events=events)
        return var


def get_parameter_args(start_time: datetime, stop_time: datetime, product: str, **kwargs) -> Dict:
    """Get parameter arguments

    Parameters
    ----------
    start_time: datetime
        parameter start time
    stop_time: datetime
        parameter stop time
    product: str
        product ID (xmlid)

    Returns
    -------
    dict
        parameter arguments in dictionary
    """
    return {'path': f"amda/{product}", 'start_time': f'{start_time.isoformat()}',
            'stop_time': f'{stop_time.isoformat()}'}



# -*- coding: utf-8 -*-
"""
.. testsetup:: *

   import speasy as spz

"""

__author__ = """Alexis Jeandet"""
__email__ = 'alexis.jeandet@member.fsf.org'
__version__ = '1.0.5'
__all__ = ['amda', 'cda', 'ssc', 'csa', 'get_data', 'SpeasyVariable', 'Catalog', 'Event', 'Dataset', 'TimeTable']
__docformat__ = "numpy"

from speasy.core.inventory.indexes import SpeasyIndex
from .products import SpeasyVariable, Catalog, Event, Dataset, TimeTable, MaybeAnyProduct
from typing import List
from .core.requests_scheduling.request_dispatch import get_data, list_providers, amda, cda, csa, ssc


# @TODO implement me, this function should be able to look inside all servers
# and return something that could be passed to get_data
def find_product(name: str) -> List[str]:
    pass


def update_inventories():
    from .core.dataprovider import PROVIDERS
    for provider in PROVIDERS.values():
        provider.update_inventory()



from typing import Optional
import json

__INDEXES_TYPES__ = {}


class SpeasyIndex:
    def __init_subclass__(cls, **kwargs):
        super().__init_subclass__(**kwargs)
        __INDEXES_TYPES__[cls.__name__] = cls

    def __init__(self, name: str, provider: str, uid: str, meta: Optional[dict] = None):
        if meta:
            self.__dict__.update(meta)
        self.__spz_provider__ = provider
        self.__spz_name__ = name
        self.__spz_uid__ = uid
        self.__spz_type__ = self.__class__.__name__

    def __eq__(self, other):
        return self.__dict__ == other.__dict__

    def clear(self):
        keys = list(self.__dict__.keys())
        for key in keys:
            if not key.startswith('__spz_'):
                self.__dict__.pop(key)

    def spz_provider(self):
        return self.__spz_provider__

    def spz_name(self):
        return self.__spz_name__

    def spz_uid(self):
        return self.__spz_uid__

    def spz_type(self):
        return self.__spz_type__

    def __repr__(self):
        return f'<SpeasyIndex: {self.spz_name()}>'


class TimetableIndex(SpeasyIndex):
    def __init__(self, name: str, provider: str, uid: str, meta: Optional[dict] = None):
        super().__init__(name, provider, uid, meta)

    def __repr__(self):
        return f'<TimetableIndex: {self.spz_name()}>'


class CatalogIndex(SpeasyIndex):
    def __init__(self, name: str, provider: str, uid: str, meta: Optional[dict] = None):
        super().__init__(name, provider, uid, meta)

    def __repr__(self):
        return f'<CatalogIndex: {self.spz_name()}>'


class ComponentIndex(SpeasyIndex):
    def __init__(self, name: str, provider: str, uid: str, meta: Optional[dict] = None):
        super().__init__(name, provider, uid, meta)

    def __repr__(self):
        return f'<ComponentIndex: {self.spz_name()}>'


class ParameterIndex(SpeasyIndex):
    def __init__(self, name: str, provider: str, uid: str, meta: Optional[dict] = None):
        super().__init__(name, provider, uid, meta)

    def __repr__(self):
        return f'<ParameterIndex: {self.spz_name()}>'

    def __iter__(self):
        return [v for v in self.__dict__.values() if type(v) is ComponentIndex].__iter__()

    def __contains__(self, item: str or ComponentIndex):
        if type(item) is ComponentIndex:
            item = item.name()
        return item in self.__dict__


class DatasetIndex(SpeasyIndex):
    def __init__(self, name: str, provider: str, uid: str, meta: Optional[dict] = None):
        super().__init__(name, provider, uid, meta)

    def __repr__(self):
        return f'<DatasetIndex: {self.spz_name()}>'

    def __iter__(self):
        return [v for v in self.__dict__.values() if type(v) is ParameterIndex].__iter__()

    def __contains__(self, item: str or ParameterIndex):
        if type(item) is ParameterIndex:
            item = item.name()
        return item in self.__dict__


def to_dict(inventory_tree: SpeasyIndex or str):
    if isinstance(inventory_tree, SpeasyIndex):
        return {key: to_dict(value) for key, value in inventory_tree.__dict__.items()}
    elif type(inventory_tree) is not str:
        return str(inventory_tree)
    return inventory_tree


def from_dict(inventory_tree: dict or str):
    if type(inventory_tree) is str:
        return inventory_tree
    idx_type = inventory_tree.pop("__spz_type__")
    idx_name = inventory_tree.pop("__spz_name__")
    idx_provider = inventory_tree.pop("__spz_provider__")
    idx_uid = inventory_tree.pop("__spz_uid__")
    idx_meta = {key: from_dict(value) for key, value in inventory_tree.items()}
    root = __INDEXES_TYPES__.get(idx_type, SpeasyIndex)(name=idx_name, provider=idx_provider, uid=idx_uid,
                                                        meta=idx_meta)
    return root


def to_json(inventory_tree: SpeasyIndex, sort_keys=True):
    return json.dumps(to_dict(inventory_tree), sort_keys=sort_keys)


def from_json(inventory_tree: str):
    return from_dict(json.loads(inventory_tree))


def make_inventory_node(parent, ctor, name, provider, uid, **meta):
    if name not in parent.__dict__:
        parent.__dict__[name] = ctor(name=name, provider=provider, uid=uid, meta=meta)
    return parent.__dict__[name]


def inventory_has_changed(orig, new):
    if orig.__dict__.keys() != new.__dict__.keys():
        return True
    for orig_key, orig_value in orig.__dict__.items():
        if orig_key != 'build_date':
            if orig_key not in new.__dict__:
                return True
            if orig.__dict__[orig_key] != new.__dict__[orig_key]:
                return True
    return False



from ._xml_catalogs_parser import load_xml_catalog
from ._cdf_masters_parser import update_tree
from ....core.index import index
from speasy.core import http 
from ....core.inventory.indexes import SpeasyIndex, to_dict, from_dict
from ....config import cdaweb as cda_cfg
from tempfile import NamedTemporaryFile
import tarfile
import os
from glob import glob

_MASTERS_CDF_PATH = f"{cda_cfg.inventory_data_path()}/masters_cdf/"
_XML_CATALOG_PATH = f"{cda_cfg.inventory_data_path()}/all.xml"


def _ensure_path_exists(path: str):
    dirname = os.path.dirname(path)
    if not os.path.exists(dirname):
        os.makedirs(dirname)


def _clean_master_cdf_folder():
    _ensure_path_exists(_MASTERS_CDF_PATH)
    cdf_files = glob(f"{_MASTERS_CDF_PATH}/*.cdf")
    for cdf_file in cdf_files:
        os.remove(cdf_file)


def _download_and_extract_master_cdf(masters_url: str):
    with NamedTemporaryFile('wb') as master_archive:
        master_archive.write(http.get(masters_url).content)
        master_archive.flush()
        tar = tarfile.open(master_archive.name)
        tar.extractall(_MASTERS_CDF_PATH)


def update_master_cdf(masters_url: str = "https://spdf.gsfc.nasa.gov/pub/software/cdawlib/0MASTERS/master.tar"):
    last_modified = http.get(masters_url, head_only=True).headers['last-modified']
    if index.get("cdaweb-inventory", "masters-last-modified", "") != last_modified:
        _clean_master_cdf_folder()
        _download_and_extract_master_cdf(masters_url)
        index.set("cdaweb-inventory", "masters-last-modified", last_modified)
        return True
    return False


def update_xml_catalog(xml_catalog_url: str = "https://spdf.gsfc.nasa.gov/pub/catalogs/all.xml"):
    last_modified = http.get(xml_catalog_url, head_only=True).headers['last-modified']
    if index.get("cdaweb-inventory", "xml_catalog-last-modified", "") != last_modified:
        _ensure_path_exists(_XML_CATALOG_PATH)
        with open(_XML_CATALOG_PATH, 'w') as f:
            f.write(http.get(xml_catalog_url).text)
            index.set("cdaweb-inventory", "xml_catalog-last-modified", last_modified)
            return True
    return False


def build_inventory(root: SpeasyIndex = None, xml_catalog_url: str = "https://spdf.gsfc.nasa.gov/pub/catalogs/all.xml",
                    masters_url: str = "https://spdf.gsfc.nasa.gov/pub/software/cdawlib/0MASTERS/master.tar"):
    needs_rebuild = update_xml_catalog(xml_catalog_url)
    needs_rebuild |= update_master_cdf(masters_url)
    if needs_rebuild or not index.contains("cdaweb-inventory", "tree"):
        root = load_xml_catalog(xml_file_path=_XML_CATALOG_PATH, root=root)
        update_tree(root=root, master_cdf_dir=_MASTERS_CDF_PATH)
        index.set("cdaweb-inventory", "tree", to_dict(root))
    else:
        t = from_dict(index.get("cdaweb-inventory", "tree"))
        root.__dict__ = t.__dict__
    return root



from typing import Union

import diskcache as dc
from .version import str_to_version, version_to_str, Version
from speasy.config import cache as cache_cfg
from contextlib import ExitStack

cache_version = str_to_version("2.0")


class CacheItem:
    def __init__(self, data, version):
        self.data = data
        self.version = version


class Cache:
    __slots__ = ['cache_file', '_data', '_hit', '_miss', 'cache_type']

    def __init__(self, cache_path: str = "", cache_type='Cache'):
        cache_path = f"{cache_path}/{cache_type}"
        if cache_type == 'Fanout':
            self._data = dc.FanoutCache(cache_path, shards=8, size_limit=cache_cfg.size())
        elif cache_type == 'Cache':
            self._data = dc.Cache(cache_path, size_limit=cache_cfg.size())
        else:
            raise ValueError(f"Unimplemented cache type: {cache_type}")

        self.cache_type = cache_type
        self._hit = 0
        self._miss = 0
        if self.version < cache_version:
            self._data.clear()
            self.version = cache_version

    @property
    def version(self):
        return str_to_version(self._data.get("cache/version", default="0.0.0"))

    @version.setter
    def version(self, v: Union[str, Version]):
        self._data["cache/version"] = v if type(v) is str else version_to_str(v)

    def disk_size(self):
        return self._data.volume()

    def stats(self):
        return {
            "hit": self._hit,
            "misses": self._miss
        }

    def __len__(self):
        return len(self._data)

    def __del__(self):
        pass

    def keys(self):
        return list(self._data)

    def __contains__(self, item):
        if item in self._data:
            self._hit += 1
            return True
        self._miss += 1
        return False

    def __getitem__(self, key):
        if key in self._data:
            self._hit += 1
        else:
            self._miss += 1
        return self._data[key]

    def __setitem__(self, key, value):
        self._data[key] = value

    def set(self, key, value, expire=None):
        self._data.set(key, value, expire=expire)

    def get(self, key, default_value=None):
        return self._data.get(key, default_value)

    def transact(self):
        if self.cache_type != 'Fanout':
            return self._data.transact()
        else:
            return ExitStack()



import pyistp
from ...products import SpeasyVariable, VariableAxis, VariableTimeAxis, DataContainer


def _fix_value_type(value):
    if type(value) in (str, int, float):
        return value
    if type(value) is list:
        return [_fix_value_type(sub_v) for sub_v in value]
    return str(value)


def _fix_attributes_types(attributes: dict):
    cleaned = {}
    for key, value in attributes.items():
        cleaned[key] = _fix_value_type(value)
    return cleaned


def _make_axis(axis, time_axis_name):
    if axis.attributes.get('DEPEND_0', '') == time_axis_name:
        is_time_dependent = True
    else:
        is_time_dependent = False
    return VariableAxis(values=axis.values.copy(), meta=_fix_attributes_types(axis.attributes), name=axis.name,
                        is_time_dependent=is_time_dependent)


def load_variable(variable="", file=None, buffer=None) -> SpeasyVariable or None:
    istp = pyistp.load(file=file, buffer=buffer)
    if istp:
        if variable in istp.data_variables():
            var = istp.data_variable(variable)
        elif variable.replace('-', '_') in istp.data_variables():  # THX CSA/ISTP
            var = istp.data_variable(variable.replace('-', '_'))
        elif variable.replace('/', '$') in istp.data_variables():  # CDA
            var = istp.data_variable(variable.replace('/', '$'))
        else:
            return None
        if var:
            time_axis_name = var.axes[0].name
            return SpeasyVariable(
                axes=[VariableTimeAxis(values=var.axes[0].values.copy(),
                                       meta=_fix_attributes_types(var.axes[0].attributes))] + [
                         _make_axis(axis, time_axis_name) for axis in var.axes[1:]],
                values=DataContainer(values=var.values.copy(), meta=_fix_attributes_types(var.attributes),
                                     name=var.name,
                                     is_time_dependent=True),
                columns=var.labels)
    return None
